# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Workflow executor provider."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Dict, Tuple

from renku.core.models.workflow.plan import AbstractPlan


class IWorkflowProvider(metaclass=ABCMeta):
    """Abstract class for executing ``Plan``."""

    @abstractmethod
    def workflow_provider(self) -> Tuple[IWorkflowProvider, str]:
        """Supported workflow description formats.

        :returns: a tuple of ``self`` and format.
        """
        pass

    @abstractmethod
    def workflow_execute(self, workflow: AbstractPlan, basedir: Path, config: Dict[str, Any]):
        """Executes a given ``AbstractPlan`` using the provider.

        :returns: a list of output paths that were generated by this workflow.
        """
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2020-2021 -Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test utility functions."""
import contextlib
import os
import traceback
import uuid
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from typing import List, Optional, Union

import pytest
from flaky import flaky
from git import Repo

from renku.core.management.command_builder.command import inject, remove_injector
from renku.core.management.dataset.datasets_provenance import DatasetsProvenance
from renku.core.management.interface.database_dispatcher import IDatabaseDispatcher
from renku.core.management.interface.dataset_gateway import IDatasetGateway
from renku.core.models.dataset import Dataset
from renku.core.models.entity import Entity
from renku.core.models.provenance.activity import Activity, Association, Generation, Usage
from renku.core.models.workflow.plan import Plan


def raises(error):
    """Wrapper around pytest.raises to support None."""
    if error:
        return pytest.raises(error)
    else:

        @contextmanager
        def not_raises():
            try:
                yield
            except Exception as e:
                raise e

        return not_raises()


def make_dataset_add_payload(project_id, urls, name=None):
    """Make dataset add request payload."""
    files = []
    for url in urls:
        if isinstance(url, tuple):
            files.append({url[0]: url[1]})

        if isinstance(url, str):
            files.append({"file_url": url})

    return {
        "project_id": project_id,
        "name": name or uuid.uuid4().hex,
        "create_dataset": True,
        "force": False,
        "files": files,
    }


def assert_dataset_is_mutated(old: Dataset, new: Dataset, mutator=None):
    """Check metadata is updated correctly after dataset mutation."""
    assert old.name == new.name
    assert old.initial_identifier == new.initial_identifier
    assert old.id != new.id
    assert old.identifier != new.identifier
    assert old.id == new.derived_from.url_id
    assert old.date_created != new.date_created
    assert new.same_as is None
    assert new.date_published is None
    assert new.identifier in new.id

    if mutator:
        old_creators = {c.email for c in old.creators}
        new_creators = {c.email for c in new.creators}
        assert new_creators == old_creators | {mutator.email}, f"{new_creators} {old_creators}"


@contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.
    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


def format_result_exception(result):
    """Format a `runner.invoke` exception result into a nice string repesentation."""

    if getattr(result, "exc_info", None):
        stacktrace = "".join(traceback.format_exception(*result.exc_info))
    else:
        stacktrace = ""

    return f"Stack Trace:\n{stacktrace}\n\nOutput:\n{result.output}"


def load_dataset(name: str) -> Optional[Dataset]:
    """Load dataset from disk."""
    datasets_provenance = DatasetsProvenance()

    return datasets_provenance.get_by_name(name)


@contextmanager
@inject.autoparams("dataset_gateway", "database_dispatcher")
def with_dataset(
    client,
    name: str = None,
    dataset_gateway: IDatasetGateway = None,
    database_dispatcher: IDatabaseDispatcher = None,
    commit_database: bool = False,
):
    """Yield an editable metadata object for a dataset."""
    dataset = client.get_dataset(name=name, strict=True, immutable=True)
    dataset._v_immutable = False

    yield dataset

    if commit_database:
        dataset_gateway.add_or_remove(dataset)
        database_dispatcher.current_database.commit()


def retry_failed(fn=None, extended: bool = False):
    """
    Decorator to run flaky with same number of max and min repetitions across all tests.

    :param extended: allow more repetitions than usual.
    """

    def decorate(fn):
        limit = 20 if extended else 5

        @flaky(max_runs=limit, min_passes=1)
        @wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    if fn:
        return decorate(fn)
    return decorate


@contextlib.contextmanager
def injection_manager(bindings):
    """Context manager to temporarly do injections."""

    def _bind(binder):
        for key, value in bindings["bindings"].items():
            binder.bind(key, value)
        for key, value in bindings["constructor_bindings"].items():
            binder.bind_to_constructor(key, value)

        return binder

    inject.configure(_bind, bind_in_runtime=False)
    try:
        yield
    finally:
        try:
            if IDatabaseDispatcher in bindings["bindings"]:
                bindings["bindings"][IDatabaseDispatcher].finalize_dispatcher()
        finally:
            remove_injector()


def write_and_commit_file(repo: Repo, path: Union[Path, str], content: str):
    """Write content to a given file and make a commit."""
    filepath = Path(path)
    path = str(path)

    filepath.parent.mkdir(parents=True, exist_ok=True)
    filepath.write_text(content)

    repo.git.add(path)
    repo.index.commit(f"Updated '{path}'")


def create_dummy_activity(
    plan: Union[Plan, str],
    usages: List[Union[Path, str]] = (),
    generations: List[Union[Path, str]] = (),
    ended_at_time=None,
) -> Activity:
    """Create a dummy activity."""
    if not isinstance(plan, Plan):
        assert isinstance(plan, str)
        plan = Plan(id=Plan.generate_id(), name=plan)

    ended_at_time = ended_at_time or (datetime.utcnow() - timedelta(hours=1))
    checksum = "abc123"

    activity_id = Activity.generate_id()

    return Activity(
        id=activity_id,
        ended_at_time=ended_at_time,
        association=Association(id=Association.generate_id(activity_id), plan=plan),
        generations=[
            Generation(
                id=Generation.generate_id(activity_id),
                entity=Entity(id=Entity.generate_id(checksum, g), checksum=checksum, path=g),
            )
            for g in generations
        ],
        usages=[
            Usage(
                id=Usage.generate_id(activity_id),
                entity=Entity(id=Entity.generate_id(checksum, u), checksum=checksum, path=u),
            )
            for u in usages
        ],
    )



# -*- coding: utf-8 -*-
#
# Copyright 2020-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service dataset jobs tests."""
import json
import tempfile
import uuid

import jwt
import pytest
from git import Repo
from werkzeug.utils import secure_filename

from renku.core.errors import DatasetExistsError, MigrationRequired, ParameterError
from renku.service.jobs.cleanup import cache_project_cleanup
from renku.service.jobs.datasets import dataset_add_remote_file, dataset_import
from renku.service.serializers.headers import JWT_TOKEN_SECRET, encode_b64
from renku.service.utils import make_project_path
from tests.service.views.test_dataset_views import assert_rpc_response
from tests.utils import retry_failed


@pytest.mark.parametrize(
    "url",
    [
        "https://dev.renku.ch/datasets/428c3626-1c56-463d-8753-336470cc6917/",
        "https://dev.renku.ch/datasets/0d4630e5-6ee5-42b3-bb3a-b3222da811ec",
    ],
)
@pytest.mark.integration
@retry_failed
def test_dataset_url_import_job(url, svc_client_with_repo):
    """Test dataset import via url."""
    svc_client, headers, project_id, url_components = svc_client_with_repo

    decoded = jwt.decode(headers["Renku-User"], JWT_TOKEN_SECRET, algorithms=["HS256"], audience="renku")
    user_data = {
        "fullname": decoded["name"],
        "email": decoded["email"],
        "user_id": encode_b64(secure_filename(decoded["sub"])),
        "token": headers["Authorization"].split("Bearer ")[-1],
    }

    payload = {
        "project_id": project_id,
        "dataset_uri": url,
    }

    response = svc_client.post("/datasets.import", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"job_id", "created_at"} == set(response.json["result"].keys())

    dest = make_project_path(
        user_data,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json["result"]["job_id"]

    commit_message = "service: import remote dataset"
    dataset_import(user_data, job_id, project_id, url, commit_message=commit_message)

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha
    assert commit_message == new_commit.message

    response = svc_client.get(f"/jobs/{job_id}", headers=headers)

    assert response
    assert_rpc_response(response)
    assert "COMPLETED" == response.json["result"]["state"]


@pytest.mark.parametrize("doi", ["10.5281/zenodo.3239980", "10.5281/zenodo.3188334", "10.7910/DVN/TJCLKP"])
@pytest.mark.integration
@pytest.mark.service
@retry_failed
def test_dataset_import_job(doi, svc_client_with_repo):
    """Test dataset import via doi."""
    svc_client, headers, project_id, url_components = svc_client_with_repo

    user_id = encode_b64(secure_filename("9ab2fc80-3a5c-426d-ae78-56de01d214df"))
    user = {"user_id": user_id}

    payload = {
        "project_id": project_id,
        "dataset_uri": doi,
    }
    response = svc_client.post("/datasets.import", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"job_id", "created_at"} == set(response.json["result"].keys())

    dest = make_project_path(
        user,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json["result"]["job_id"]

    commit_message = "service: import remote dataset"
    dataset_import(user, job_id, project_id, doi, commit_message=commit_message)

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha
    assert commit_message == new_commit.message

    response = svc_client.get(f"/jobs/{job_id}", headers=headers)
    assert response
    assert_rpc_response(response)
    assert "COMPLETED" == response.json["result"]["state"]


@pytest.mark.parametrize(
    "doi,expected_err",
    [
        # not valid doi
        ("junkjunkjunk", "Invalid parameter value"),
        # not existing doi
        ("10.5281/zenodo.11111111111111111", "Invalid parameter value"),
    ],
)
@pytest.mark.integration
@pytest.mark.service
@retry_failed
def test_dataset_import_junk_job(doi, expected_err, svc_client_with_repo):
    """Test dataset import."""
    svc_client, headers, project_id, url_components = svc_client_with_repo

    user_id = encode_b64(secure_filename("9ab2fc80-3a5c-426d-ae78-56de01d214df"))
    user = {"user_id": user_id}

    payload = {
        "project_id": project_id,
        "dataset_uri": doi,
    }
    response = svc_client.post("/datasets.import", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"job_id", "created_at"} == set(response.json["result"].keys())

    dest = make_project_path(
        user,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json["result"]["job_id"]

    with pytest.raises(ParameterError):
        dataset_import(user, job_id, project_id, doi)

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha == new_commit.hexsha

    response = svc_client.get(f"/jobs/{job_id}", data=json.dumps(payload), headers=headers)

    assert_rpc_response(response)
    extras = response.json["result"]["extras"]

    assert "error" in extras
    assert expected_err in extras["error"]


@pytest.mark.parametrize("doi", ["10.5281/zenodo.3634052"])
@pytest.mark.integration
@pytest.mark.service
@retry_failed
def test_dataset_import_twice_job(doi, svc_client_with_repo):
    """Test dataset import."""
    svc_client, headers, project_id, url_components = svc_client_with_repo

    user_id = encode_b64(secure_filename("9ab2fc80-3a5c-426d-ae78-56de01d214df"))
    user = {"user_id": user_id}

    payload = {
        "project_id": project_id,
        "dataset_uri": doi,
    }
    response = svc_client.post("/datasets.import", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"job_id", "created_at"} == set(response.json["result"].keys())

    dest = make_project_path(
        user,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )

    old_commit = Repo(dest).head.commit
    job_id = response.json["result"]["job_id"]

    dataset_import(user, job_id, project_id, doi)

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha != new_commit.hexsha

    with pytest.raises(DatasetExistsError):
        dataset_import(user, job_id, project_id, doi)

    new_commit2 = Repo(dest).head.commit
    assert new_commit.hexsha == new_commit2.hexsha

    response = svc_client.get(f"/jobs/{job_id}", data=json.dumps(payload), headers=headers)

    assert_rpc_response(response)
    extras = response.json["result"]["extras"]

    assert "error" in extras
    assert "Dataset exists" in extras["error"]


@pytest.mark.parametrize(
    "url", ["https://gist.github.com/jsam/d957f306ed0fe4ff018e902df6a1c8e3", "https://tinyurl.com/y6gne4ct"]
)
@pytest.mark.integration
@pytest.mark.service
@retry_failed
def test_dataset_add_remote_file(url, svc_client_with_repo):
    """Test dataset add a remote file."""
    svc_client, headers, project_id, url_components = svc_client_with_repo

    user_id = encode_b64(secure_filename("9ab2fc80-3a5c-426d-ae78-56de01d214df"))
    user = {"user_id": user_id}

    payload = {"project_id": project_id, "name": uuid.uuid4().hex, "create_dataset": True, "files": [{"file_url": url}]}
    response = svc_client.post("/datasets.add", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"files", "name", "project_id", "remote_branch"} == set(response.json["result"].keys())

    dest = make_project_path(
        user,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )
    old_commit = Repo(dest).head.commit
    job_id = response.json["result"]["files"][0]["job_id"]
    commit_message = "service: dataset add remote file"

    dataset_add_remote_file(user, job_id, project_id, True, commit_message, payload["name"], url)

    new_commit = Repo(dest).head.commit

    assert old_commit.hexsha != new_commit.hexsha
    assert commit_message == new_commit.message


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_add_file_job(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Add a file to a new dataset on a remote repository."""
    from renku.service.serializers.datasets import DatasetAddRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    _, _, cache = svc_client_cache
    user = cache.ensure_user(view_user_data)
    renku_module = "renku.service.controllers.datasets_add_file"
    renku_ctrl = "DatasetsAddFileCtrl"

    context = DatasetAddRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": uuid.uuid4().hex,
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
            "migrate_project": True,
            "create_dataset": True,
            "files": [{"file_url": "https://tinyurl.com/y6gne4ct"}],
        }
    )

    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    updated_job = delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"remote_branch", "project_id", "files", "name"} == updated_job.ctrl_result["result"].keys()


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_add_file_job_failure(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Add a file to a new dataset on a remote repository."""
    from renku.service.serializers.datasets import DatasetAddRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch
    _, _, cache = svc_client_cache

    view_user_data["user_id"] = uuid.uuid4().hex
    user = cache.ensure_user(view_user_data)
    renku_module = "renku.service.controllers.datasets_add_file"
    renku_ctrl = "DatasetsAddFileCtrl"

    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.write(b"data123")
    temp_file.close()

    context = DatasetAddRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": uuid.uuid4().hex,
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
            "migrate_project": False,
            "create_dataset": True,
            "files": [{"file_path": temp_file.name}],
        }
    )

    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    with pytest.raises(MigrationRequired):
        delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)


@pytest.mark.parametrize("doi", ["10.5281/zenodo.3761586"])
@pytest.mark.integration
@pytest.mark.service
def test_dataset_project_lock(doi, svc_client_with_repo):
    """Test dataset project lock."""
    svc_client, headers, project_id, url_components = svc_client_with_repo
    user_id = encode_b64(secure_filename("9ab2fc80-3a5c-426d-ae78-56de01d214df"))
    user = {"user_id": user_id}

    payload = {
        "project_id": project_id,
        "dataset_uri": doi,
    }
    response = svc_client.post("/datasets.import", data=json.dumps(payload), headers=headers)

    assert response
    assert_rpc_response(response)
    assert {"job_id", "created_at"} == set(response.json["result"].keys())

    dest = make_project_path(
        user,
        {
            "owner": url_components.owner,
            "name": url_components.name,
            "slug": url_components.slug,
            "project_id": project_id,
        },
    )

    old_commit = Repo(dest).head.commit

    cache_project_cleanup()

    new_commit = Repo(dest).head.commit
    assert old_commit.hexsha == new_commit.hexsha
    assert dest.exists() and [file for file in dest.glob("*")]


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_create_dataset_job(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Create a new dataset successfully."""
    from renku.service.serializers.datasets import DatasetCreateRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetCreateRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": uuid.uuid4().hex,
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
            "migrate_project": True,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_create"
    renku_ctrl = "DatasetsCreateCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    updated_job = delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"name", "remote_branch"} == updated_job.ctrl_result["result"].keys()


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_create_dataset_failure(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Create a new dataset successfully."""
    from renku.service.serializers.datasets import DatasetCreateRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetCreateRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": uuid.uuid4().hex,
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_create"
    renku_ctrl = "DatasetsCreateCtrl"

    view_user_data["user_id"] = uuid.uuid4().hex
    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    with pytest.raises(MigrationRequired):
        delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_remove_dataset_job(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Create a dataset was removed successfully."""
    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job
    from renku.service.serializers.datasets import DatasetRemoveRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch
    _, _, cache = svc_client_cache
    user = cache.ensure_user(view_user_data)

    request_payload = {
        "git_url": it_remote_repo_url,
        "ref": branch,
        "name": "mydata",
        "migrate_project": True,
    }

    context = DatasetRemoveRequest().load(request_payload)
    renku_module = "renku.service.controllers.datasets_remove"
    renku_ctrl = "DatasetsRemoveCtrl"

    delete_job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    updated_job = delayed_ctrl_job(context, view_user_data, delete_job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"name", "remote_branch"} == updated_job.ctrl_result["result"].keys()


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_remove_dataset_job_failure(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Create a dataset was removed successfully."""
    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job
    from renku.service.serializers.datasets import DatasetRemoveRequest

    it_remote_repo_url, ref = it_remote_repo_url_temp_branch
    _, _, cache = svc_client_cache
    user = cache.ensure_user(view_user_data)
    dataset_name = uuid.uuid4().hex

    request_payload = {
        "git_url": it_remote_repo_url,
        "ref": ref,
        "name": dataset_name,
    }

    context = DatasetRemoveRequest().load(request_payload)
    renku_module = "renku.service.controllers.datasets_remove"
    renku_ctrl = "DatasetsRemoveCtrl"

    delete_job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    with pytest.raises(MigrationRequired):
        delayed_ctrl_job(context, view_user_data, delete_job.job_id, renku_module, renku_ctrl)


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_edit_dataset_job(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Edit a dataset successfully."""
    from renku.service.serializers.datasets import DatasetEditRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetEditRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": "mydata",
            "title": f"new title => {uuid.uuid4().hex}",
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
            "migrate_project": True,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_edit"
    renku_ctrl = "DatasetsEditCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    updated_job = delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"warnings", "remote_branch", "edited"} == updated_job.ctrl_result["result"].keys()
    assert {"title"} == updated_job.ctrl_result["result"]["edited"].keys()


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_edit_dataset_job_failure(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Edit a dataset with a failure."""
    from renku.service.serializers.datasets import DatasetEditRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetEditRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": "mydata",
            "title": f"new title => {uuid.uuid4().hex}",
            "migrate_project": False,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_edit"
    renku_ctrl = "DatasetsEditCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    with pytest.raises(MigrationRequired):
        delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_unlink_dataset_job(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Unlink a file from a dataset successfully."""
    from renku.service.serializers.datasets import DatasetUnlinkRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetUnlinkRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": "ds1",
            "include_filters": ["data1"],
            # NOTE: We test with this only to check that recursive invocation is being prevented.
            "is_delayed": True,
            "migrate_project": True,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_unlink"
    renku_ctrl = "DatasetsUnlinkCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    updated_job = delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"unlinked", "remote_branch"} == updated_job.ctrl_result["result"].keys()
    assert ["data/data/data1"] == updated_job.ctrl_result["result"]["unlinked"]


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_delay_unlink_dataset_job_failure(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Unlink a file from a dataset failure."""
    from renku.service.serializers.datasets import DatasetUnlinkRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetUnlinkRequest().load(
        {"git_url": it_remote_repo_url, "ref": branch, "name": "ds1", "include_filters": ["data1"]}
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_unlink"
    renku_ctrl = "DatasetsUnlinkCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    with pytest.raises(MigrationRequired):
        delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_unlink_dataset_sync(svc_client_cache, it_remote_repo_url_temp_branch, view_user_data):
    """Unlink a file from a dataset successfully."""
    from renku.service.serializers.datasets import DatasetUnlinkRequest

    it_remote_repo_url, branch = it_remote_repo_url_temp_branch

    context = DatasetUnlinkRequest().load(
        {
            "git_url": it_remote_repo_url,
            "ref": branch,
            "name": "ds1",
            "include_filters": ["data1"],
            "migrate_project": True,
        }
    )

    _, _, cache = svc_client_cache
    renku_module = "renku.service.controllers.datasets_unlink"
    renku_ctrl = "DatasetsUnlinkCtrl"

    user = cache.ensure_user(view_user_data)
    job = cache.make_job(
        user, job_data={"ctrl_context": {**context, "renku_module": renku_module, "renku_ctrl": renku_ctrl}}
    )

    from renku.service.jobs.delayed_ctrl import delayed_ctrl_job

    updated_job = delayed_ctrl_job(context, view_user_data, job.job_id, renku_module, renku_ctrl)

    assert updated_job
    assert {"unlinked", "remote_branch"} == updated_job.ctrl_result["result"].keys()
    assert ["data/data/data1"] == updated_job.ctrl_result["result"]["unlinked"]



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder."""

import contextlib
import functools
import threading
import typing
from collections import defaultdict

import click
import inject

from renku.core import errors
from renku.core.utils.communication import CommunicationCallback

_LOCAL = threading.local()


def check_finalized(f):
    """Decorator to prevent modification of finalized builders."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """Decorator to prevent modification of finalized builders."""
        if not args or not isinstance(args[0], Command):
            raise errors.ParameterError("Command hooks need to be `Command` object methods.")

        if args[0].finalized:
            raise errors.CommandFinalizedError("Cannot modify a finalized `Command`.")

        return f(*args, **kwargs)

    return wrapper


def _patched_get_injector_or_die() -> inject.Injector:
    """Patched version of get_injector_or_die with thread local injectors.

    Allows deferred definition of an injector per thread.
    """
    injector = getattr(_LOCAL, "injector", None)
    if not injector:
        raise inject.InjectorException("No injector is configured")

    return injector


def _patched_configure(
    config: typing.Optional[inject.BinderCallable] = None, bind_in_runtime: bool = True
) -> inject.Injector:
    """Create an injector with a callable config or raise an exception when already configured."""

    if getattr(_LOCAL, "injector", None):
        raise inject.InjectorException("Injector is already configured")

    _LOCAL.injector = inject.Injector(config, bind_in_runtime=bind_in_runtime)

    return _LOCAL.injector


inject.configure = _patched_configure
inject.get_injector_or_die = _patched_get_injector_or_die


def remove_injector():
    """Remove a thread-local injector."""
    if getattr(_LOCAL, "injector", None):
        del _LOCAL.injector


@contextlib.contextmanager
def replace_injection(bindings, constructor_bindings=None):
    """Temporarily inject various test objects."""
    constructor_bindings = constructor_bindings or {}

    def bind(binder):
        for key, value in bindings.items():
            binder.bind(key, value)
        for key, value in constructor_bindings.items():
            binder.bind_to_constructor(key, value)

    old_injector = getattr(_LOCAL, "injector", None)
    try:
        if old_injector:
            remove_injector()
        inject.configure(bind, bind_in_runtime=False)

        yield
    finally:
        remove_injector()

        if old_injector:
            _LOCAL.injector = old_injector


def update_injected_client(new_client, update_database: bool = True):
    """Update the injected client instance.

    Necessary because we sometimes use attr.evolve to modify a client and this doesn't affect the injected instance.
    """
    from renku.core.management import LocalClient
    from renku.core.metadata.database import Database

    injector = getattr(_LOCAL, "injector", None)

    if not injector:
        raise inject.InjectorException("No injector is configured")

    injector._bindings[LocalClient] = lambda: new_client
    injector._bindings["LocalClient"] = lambda: new_client

    if update_database and Database in injector._bindings:
        database = Database.from_path(path=new_client.database_path)
        injector._bindings[Database] = lambda: database


class Command:
    """Base renku command builder."""

    CLIENT_HOOK_ORDER = 1

    def __init__(self) -> None:
        """__init__ of Command."""
        self.injection_pre_hooks = defaultdict(list)
        self.pre_hooks = defaultdict(list)
        self.post_hooks = defaultdict(list)
        self._operation = None
        self._finalized = False
        self._track_std_streams = False
        self._working_directory = None

    def __getattr__(self, name: str) -> typing.Any:
        """Bubble up attributes of wrapped builders."""
        if "_builder" in self.__dict__:
            return getattr(self._builder, name)

        raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

    def __setattr__(self, name: str, value: typing.Any) -> None:
        """Set attributes of wrapped builders."""
        if hasattr(self, "_builder") and self.__class__ is not self._builder.__class__:
            self._builder.__setattr__(name, value)

        object.__setattr__(self, name, value)

    def _injection_pre_hook(self, builder: "Command", context: dict, *args, **kwargs) -> None:
        """Setup dependency injections."""
        from renku.core.management import LocalClient
        from renku.core.management.command_builder.client_dispatcher import ClientDispatcher
        from renku.core.management.interface.client_dispatcher import IClientDispatcher
        from renku.core.management.repository import default_path

        dispatcher = ClientDispatcher()

        ctx = click.get_current_context(silent=True)
        if ctx is None:
            dispatcher.push_client_to_stack(path=default_path(self._working_directory or "."))
            ctx = click.Context(click.Command(builder._operation))
        else:
            client = ctx.ensure_object(LocalClient)
            dispatcher.push_created_client_to_stack(client)

        context["bindings"] = {IClientDispatcher: dispatcher, "IClientDispatcher": dispatcher}
        context["constructor_bindings"] = {}
        context["client_dispatcher"] = dispatcher
        context["click_context"] = ctx

    def _pre_hook(self, builder: "Command", context: dict, *args, **kwargs) -> None:
        """Setup local client."""

        stack = contextlib.ExitStack()
        context["stack"] = stack

    def _post_hook(self, builder: "Command", context: dict, result: "CommandResult", *args, **kwargs) -> None:
        """Post-hook method."""
        remove_injector()

        if result.error:
            raise result.error

    def execute(self, *args, **kwargs) -> "CommandResult":
        """Execute the wrapped operation.

        First executes `pre_hooks` in ascending `order`, passing a read/write context between them.
        It then calls the wrapped `operation`. The result of the operation then gets pass to all the `post_hooks`,
        but in descending `order`. It then returns the result or error if there was one.
        """
        if not self.finalized:
            raise errors.CommandNotFinalizedError("Call `build()` before executing a command")

        context = {}
        if any(self.injection_pre_hooks):
            order = sorted(self.injection_pre_hooks.keys())

            for o in order:
                for hook in self.injection_pre_hooks[o]:
                    hook(self, context, *args, **kwargs)

        def _bind(binder):
            for key, value in context["bindings"].items():
                binder.bind(key, value)
            for key, value in context["constructor_bindings"].items():
                binder.bind_to_constructor(key, value)

            return binder

        inject.configure(_bind, bind_in_runtime=False)

        if any(self.pre_hooks):
            order = sorted(self.pre_hooks.keys())

            for o in order:
                for hook in self.pre_hooks[o]:
                    try:
                        hook(self, context, *args, **kwargs)
                    except (Exception, BaseException):
                        # don't leak injections from failed hook
                        remove_injector()
                        raise

        output = None
        error = None

        try:
            with context["stack"]:
                output = context["click_context"].invoke(self._operation, *args, **kwargs)
        except errors.RenkuException as e:
            error = e
        except (Exception, BaseException):
            remove_injector()
            raise

        result = CommandResult(output, error, CommandResult.FAILURE if error else CommandResult.SUCCESS)

        if any(self.post_hooks):
            order = sorted(self.post_hooks.keys(), reverse=True)

            for o in order:
                for hook in self.post_hooks[o]:
                    hook(self, context, result, *args, **kwargs)

        return result

    @property
    def finalized(self) -> bool:
        """Whether this builder is still being constructed or has been finalized."""
        if hasattr(self, "_builder"):
            return self._builder.finalized
        return self._finalized

    @check_finalized
    def add_injection_pre_hook(self, order: int, hook: typing.Callable):
        """Add a pre-execution hook for dependency injection.

        :param order: Determines the order of executed hooks, lower numbers get executed first.
        :param hook: The hook to add
        """
        if hasattr(self, "_builder"):
            self._builder.add_injection_pre_hook(order, hook)
        else:
            self.injection_pre_hooks[order].append(hook)

    @check_finalized
    def add_pre_hook(self, order: int, hook: typing.Callable):
        """Add a pre-execution hook.

        :param order: Determines the order of executed hooks, lower numbers get executed first.
        :param hook: The hook to add
        """
        if hasattr(self, "_builder"):
            self._builder.add_pre_hook(order, hook)
        else:
            self.pre_hooks[order].append(hook)

    @check_finalized
    def add_post_hook(self, order: int, hook: typing.Callable):
        """Add a post-execution hook.

        :param order: Determines the order of executed hooks, lower numbers get executed first.
        :param hook: The hook to add
        """
        if hasattr(self, "_builder"):
            self._builder.add_post_hook(order, hook)
        else:
            self.post_hooks[order].append(hook)

    @check_finalized
    def build(self) -> "Command":
        """Build (finalize) the command."""
        if not self._operation:
            raise errors.ConfigurationError("`Command` needs to have a wrapped `command` set")
        self.add_injection_pre_hook(self.CLIENT_HOOK_ORDER, self._injection_pre_hook)
        self.add_pre_hook(self.CLIENT_HOOK_ORDER, self._pre_hook)
        self.add_post_hook(self.CLIENT_HOOK_ORDER, self._post_hook)

        self._finalized = True

        return self

    @check_finalized
    def command(self, operation: typing.Callable):
        """Set the wrapped command.

        :param operation: The function to wrap in the command builder.
        """
        self._operation = operation

        return self

    @check_finalized
    def working_directory(self, directory: str) -> "Command":
        """Set the working directory for the command.

        :param directory: The working directory to work in.
        """
        self._working_directory = directory

        return self

    @check_finalized
    def track_std_streams(self) -> "Command":
        """Whether to track STD streams or not."""
        self._track_std_streams = True

        return self

    @check_finalized
    def with_git_isolation(self) -> "Command":
        """Whether to run in git isolation or not."""
        from renku.core.management.command_builder.repo import Isolation

        return Isolation(self)

    @check_finalized
    def with_commit(
        self, message: str = None, commit_if_empty: bool = False, raise_if_empty: bool = False, commit_only: bool = None
    ) -> "Command":
        """Create a commit.

        :param message: The commit message. Autogenerated if left empty.
        :param commit_if_empty: Whether to commit if there are no modified files .
        :param raise_if_empty: Whether to raise an exception if there are no modified files.
        :param commit_only: Only commit the supplied paths.
        """
        from renku.core.management.command_builder.repo import Commit

        return Commit(self, message, commit_if_empty, raise_if_empty, commit_only)

    @check_finalized
    def lock_project(self) -> "Command":
        """Acquire a lock for the whole project."""
        from renku.core.management.command_builder.lock import ProjectLock

        return ProjectLock(self)

    @check_finalized
    def lock_dataset(self) -> "Command":
        """Acquire a lock for a dataset."""
        from renku.core.management.command_builder.lock import DatasetLock

        return DatasetLock(self)

    @check_finalized
    def require_migration(self) -> "Command":
        """Check if a migration is needed."""
        from renku.core.management.command_builder.migration import RequireMigration

        return RequireMigration(self)

    @check_finalized
    def require_clean(self) -> "Command":
        """Check that the repository is clean."""
        from renku.core.management.command_builder.repo import RequireClean

        return RequireClean(self)

    @check_finalized
    def require_nodejs(self) -> "Command":
        """Ensure nodejs is installed."""
        from renku.core.management.command_builder.nodejs import RequireNodeJs

        return RequireNodeJs(self)

    @check_finalized
    def with_communicator(self, communicator: CommunicationCallback) -> "Command":
        """Create a communicator."""
        from renku.core.management.command_builder.communication import Communicator

        return Communicator(self, communicator)

    @check_finalized
    def with_database(self, write: bool = False, path: str = None, create: bool = False) -> "Command":
        """Provide an object database connection."""
        from renku.core.management.command_builder.database import DatabaseCommand

        return DatabaseCommand(self, write, path, create)


class CommandResult:
    """The result of a command.

    The return value of the command is set as `.output`, if there was an error, it is set as `.error`, and
    the status of the command is set to either `CommandResult.SUCCESS` or CommandResult.FAILURE`.
    """

    SUCCESS = 0

    FAILURE = 1

    def __init__(self, output, error, status) -> None:
        """__init__ of CommandResult."""
        self.output = output
        self.error = error
        self.status = status



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku ``rerun`` command."""

from collections import defaultdict
from typing import List, Set

from renku.core import errors
from renku.core.commands.workflow import execute_workflow
from renku.core.management.command_builder.command import Command, inject
from renku.core.management.interface.activity_gateway import IActivityGateway
from renku.core.management.interface.client_dispatcher import IClientDispatcher
from renku.core.management.workflow import sort_activities
from renku.core.models.provenance.activity import Activity
from renku.core.utils import communication
from renku.core.utils.metadata import add_activity_if_recent
from renku.core.utils.os import get_relative_paths


def rerun_command():
    """Recreate files generated by a sequence of ``run`` commands."""
    return (
        Command()
        .command(_rerun)
        .require_migration()
        .require_clean()
        .require_nodejs()
        .with_database(write=True)
        .with_commit()
    )


@inject.autoparams()
def _rerun(
    dry_run: bool,
    sources: List[str],
    paths: List[str],
    client_dispatcher: IClientDispatcher,
    activity_gateway: IActivityGateway,
):
    client = client_dispatcher.current_client

    sources = sources or []
    sources = get_relative_paths(base=client.path, paths=sources)
    paths = paths or []
    paths = get_relative_paths(base=client.path, paths=paths)

    activities = _get_activities(paths, sources, activity_gateway)

    if len(activities) == 0:
        raise errors.NothingToExecuteError()

    activities = sort_activities(activities)
    if dry_run:
        return activities, set(sources)

    plans = [a.plan_with_values for a in activities]

    execute_workflow(plans=plans, command_name="rerun")


def _get_activities(paths: List[str], sources: List[str], activity_gateway: IActivityGateway) -> Set[Activity]:
    all_activities = defaultdict(set)

    def include_newest_activity(activity):
        existing_activities = all_activities[activity.association.plan.id]
        add_activity_if_recent(activity=activity, activities=existing_activities)

    for path in paths:
        activities = activity_gateway.get_activities_by_generation(path)

        if len(activities) == 0:
            communication.warn(f"Path '{path}' is not generated by any workflows.")
            continue

        latest_activity = max(activities, key=lambda a: a.ended_at_time)

        upstream_chains = activity_gateway.get_upstream_activity_chains(latest_activity)

        if sources:
            # NOTE: Add the activity to check if it also matches the condition
            upstream_chains.append((latest_activity,))
            # NOTE: Only include paths that is using at least one of the sources
            upstream_chains = [c for c in upstream_chains if any(u.entity.path in sources for u in c[-1].usages)]

            # NOTE: Include activity only if any of its upstream match the condition
            if upstream_chains:
                include_newest_activity(latest_activity)
        else:
            include_newest_activity(latest_activity)

        for chain in upstream_chains:
            for activity in chain:
                include_newest_activity(activity)

    return {a for activities in all_activities.values() for a in activities}



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service fixtures for jobs testing."""
import pytest

from tests.utils import modified_environ


@pytest.fixture
def service_job(svc_client, mock_redis):
    """Ensure correct environment during testing of service jobs."""
    with modified_environ(RENKU_SVC_CLEANUP_TTL_FILES="0", RENKU_SVC_CLEANUP_TTL_PROJECTS="0"):
        yield svc_client, mock_redis



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``graph`` command."""

import os

import pytest

from renku.cli import cli
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from tests.utils import format_result_exception, modified_environ


@pytest.mark.parametrize("revision", [None, "HEAD", "HEAD^^", "HEAD^^..HEAD"])
@pytest.mark.parametrize("format", ["json-ld", "rdf", "nt"])
def test_graph_export_validation(runner, client, directory_tree, run, revision, format):
    """Test graph validation when exporting."""
    assert 0 == runner.invoke(cli, ["dataset", "add", "-c", "my-data", str(directory_tree)]).exit_code

    file1 = client.path / DATA_DIR / "my-data" / directory_tree.name / "file1"
    file2 = client.path / DATA_DIR / "my-data" / directory_tree.name / "dir1" / "file2"
    assert 0 == run(["run", "head", str(file1)], stdout="out1")
    assert 0 == run(["run", "tail", str(file2)], stdout="out2")

    result = runner.invoke(cli, ["graph", "export", "--format", format, "--strict", "--revision", revision])

    assert 0 == result.exit_code, format_result_exception(result)

    assert "https://localhost" in result.output
    assert "https://renkulab.io" not in result.output

    with modified_environ(RENKU_DOMAIN="https://renkulab.io"):
        result = runner.invoke(cli, ["graph", "export", "--format", format, "--strict", "--revision", revision])

        assert 0 == result.exit_code, format_result_exception(result)

        assert "https://localhost" not in result.output
        assert "https://renkulab.io" in result.output


@pytest.mark.serial
@pytest.mark.shelled
@pytest.mark.parametrize("format", ["json-ld", "nt", "rdf"])
def test_graph_export_strict_run(runner, project, run_shell, format):
    """Test graph export output of run command."""
    # Run a shell command with pipe.
    result = run_shell('renku run echo "my input string" > my_output_file')

    # Assert created output file.
    result = runner.invoke(cli, ["graph", "export", "--strict", "--format={}".format(format)])
    assert 0 == result.exit_code, format_result_exception(result)
    assert "my_output_file" in result.output
    assert "my input string" in result.output


@pytest.mark.parametrize("format", ["json-ld", "nt", "rdf"])
def test_graph_export_strict_dataset(tmpdir, runner, project, client, format, subdirectory):
    """Test output of graph export for dataset add."""
    result = runner.invoke(cli, ["dataset", "create", "my-dataset"])
    assert 0 == result.exit_code, format_result_exception(result)
    paths = []
    test_paths = []
    for i in range(3):
        new_file = tmpdir.join("file_{0}".format(i))
        new_file.write(str(i))
        paths.append(str(new_file))
        test_paths.append(os.path.relpath(str(new_file), str(project)))

    # add data
    result = runner.invoke(cli, ["dataset", "add", "my-dataset"] + paths)
    assert 0 == result.exit_code, format_result_exception(result)

    result = runner.invoke(cli, ["graph", "export", "--strict", f"--format={format}"])
    assert 0 == result.exit_code, format_result_exception(result)
    assert all(p in result.output for p in test_paths), result.output



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage the set of CWL files created by ``renku`` commands.

Manipulating workflows
~~~~~~~~~~~~~~~~~~~~~~

Listing workflows:

.. code-block:: console

    $ renku workflow ls
    26be2e8d66f74130a087642768f2cef0_rerun.yaml:
    199c4b9d462f4b27a4513e5e55f76eb2_cat.yaml:
    9bea2eccf9624de387d9b06e61eec0b6_rerun.yaml:
    b681b4e229764ceda161f6551370af12_update.yaml:
    25d0805243e3468d92a3786df782a2c4_rerun.yaml:

Each ``*.yaml`` file corresponds to a renku run/update/rerun execution.

Exporting workflows:

You can export the workflow to create a file as Common Workflow Language
by using:

.. code-block:: console

    $ renku workflow set-name create output_file
    baseCommand:
    - cat
    class: CommandLineTool
    cwlVersion: v1.0
    id: 22943eca-fa4c-4f3b-a92d-f6ac7badc0d2
    inputs:
    - default:
        class: File
        path: /home/user/project/intermediate
    id: inputs_1
    inputBinding:
        position: 1
    type: File
    - default:
        class: File
        path: /home/user/project/intermediate2
    id: inputs_2
    inputBinding:
        position: 2
    type: File
    outputs:
    - id: output_stdout
    streamable: false
    type: stdout
    requirements:
    InitialWorkDirRequirement:
        listing:
        - entry: $(inputs.inputs_1)
        entryname: intermediate
        writable: false
        - entry: $(inputs.inputs_2)
        entryname: intermediate2
        writable: false
    stdout: output_file

You can use ``--revision`` to specify the revision of the output file to
generate the workflow for. You can also export to a file directly with
``-o <path>``.


Composing workflows into larger workflows
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

For more complex workflows consisting of several steps, you can use the
``renku workflow group`` command. This creates a new workflow that has
substeps.

The basic usage is:

.. code-block:: console

   $ renku run --name step1 -- command
   $ renku run --name step2 -- command
   $ renku workflow group my-grouped-workflow step1 step2

This would create a new workflow ``my-grouped-workflow`` that consists
of ``step1`` and ``step2`` as steps. This new workflow is just
like any other workflow in renku in that it can be executed, exported
or grouped with other workflows.

You can expose parameters of child steps on the parent workflow using
``--map``/``-m``  arguments followed by a mapping expression. Mapping expressions
take the form of ``<name>=<expression>`` where ``name`` is the name of the
property to be created on the parent workflow and expression points to one
or more fields on the child steps that should be mapped to this property.
The expressions come in two flavors, absolute references using the names
of workflows and properties, and relative references specifying the
position within a workflow.

An absolute expression in the example above could be ``step1.my_dataset``
to refer to the input, output or argument named ``my_dataset` on the step
``step1``. A relative expression could be ``@step2.@output1`` to refer
to the first output of the second step of the grouped workflow.

Valid relative expressions are ``@input<n>``, ``@output<n>`` and ``@param<n>``
for the n'th input, output or argument of a step, respectively. For referring
to steps inside a grouped workflow, you can use ``@step<n>``. For referencing
a mapping on a grouped workflow, you can use ``@mapping<n>``. Of course, the
names of the objects for all these cases also work.

The expressions can also be combined using ``,`` if a mapping should point
to more than one parameter of a child step.

You can mix absolute and relative reference in the same expression, as you see fit.

A full example of this would be:

.. code-block:: console

   $ renku workflow group --map input_file=step1.@input2 \
       --map output_file=@step1.my-output,@step2.step2s_output \
       my-grouped-workflow step1 step2

This would create a mapping called ``input_file`` on the parent workflow that
points to the second input of ``step1`` and a mapping called ``output_file``
that points to both the output ``my-output`` on ``step1`` and
``step2s_output`` on ``step2``.

You can also set default values for mappings, which override the default values
of the parameters they're pointing to by using the ``--set``/``-s`` parameter, for
instance:

.. code-block:: console

   $ renku workflow group --map input_file=step1.@input2 \
       --set input_file=data.csv
       my-grouped-workflow step1 step2


This would lead to ``data.csv`` being used for the second input of
``step1`` when ``my-grouped-workflow`` is executed (if it isn't overridden
at execution time).

You can add a description to the mappings to make them more human-readable
by using the ``--describe-param``/``-p`` parameter, as shown here:



.. code-block:: console

   $ renku workflow group --map input_file=step1.@input2 \
       -p input_file="The dataset to process"
       my-grouped-workflow step1 step2

You can also expose all inputs, outputs or parameters of child steps by
using ``--map-inputs``, ``--map-outputs`` or ``--map-params``, respectively.

On execution, renku will automatically detect links between steps, if an input
of one step uses the same path as an output of another step, and execute
them in the correct order. Since this depends on what values are passed
at runtime, you might want to enforce a certain order of steps by explicitely
mapping outputs to inputs.

You can do that using the ``--link <source>=<sink>`` parameters, e.g.
``--link step1.@output1=step2.@input1``. This gets recorded on the
workflow template and forces ``step2.@input1`` to always be set to the same
path as ``step1.@output1``, irrespective of which values are passed at
execution time.

This way, you can ensure that the steps in your workflow are always executed
in the correct order and that the dependencies between steps are modelled
correctly.

Renku can also add links for you automatically based on the default values
of inputs and outputs, where inputs/outputs that have the same path get
linked in the grouped run. To do this, pass the ``--link-all`` flag.

.. warning:: Due to workflows having to be directed acyclic graphs, cycles
   in the dependencies are not allowed. E.g. step1 depending on step2
   depending on step1 is not allowed. Additionally, the flow of information
   has to be from outputs to inputs or parameters, so you cannot map an input
   to an output, only the other way around.

Values on inputs/outputs/parameters get set according to the following
order of precedence (lower precedence first):

- Default value on a input/output/parameter
- Default value on a mapping to the input/output/parameter
- Value passed to a mapping to the input/output/parameter
- Value passed to the input/output/parameter
- Value propagated to an input from the source of a workflow link

Input and output files
~~~~~~~~~~~~~~~~~~~~~~

You can list input and output files generated in the repository by running
``renku workflow inputs`` and ``renku workflow outputs`` commands. Alternatively,
you can check if all paths specified as arguments are input or output files
respectively.

.. code-block:: console

   $ renku run wc < source.txt > result.wc
   $ renku workflow inputs
   source.txt
   $ renku workflow outputs
   result.wc
   $ renku workflow outputs source.txt
   $ echo $?  # last command finished with an error code
   1

"""

from pathlib import Path

import click
from rich.console import Console
from rich.markdown import Markdown

from renku.cli.utils.callback import ClickCallback
from renku.core.commands.echo import ERROR
from renku.core.commands.format.workflow import WORKFLOW_COLUMNS, WORKFLOW_FORMATS
from renku.core.commands.view_model.composite_plan import CompositePlanViewModel
from renku.core.commands.view_model.plan import PlanViewModel
from renku.core.commands.workflow import (
    compose_workflow_command,
    edit_workflow_command,
    execute_workflow_command,
    export_workflow_command,
    list_workflows_command,
    remove_workflow_command,
    show_workflow_command,
    workflow_inputs_command,
    workflow_outputs_command,
)
from renku.core.plugins.provider import available_workflow_providers
from renku.core.plugins.workflow import supported_formats


def _print_plan(plan: PlanViewModel):
    """Print a plan to stdout."""
    click.echo(click.style("Id: ", bold=True, fg="magenta") + click.style(plan.id, bold=True))
    click.echo(click.style("Name: ", bold=True, fg="magenta") + click.style(plan.name, bold=True))

    if plan.description:
        Console().print(Markdown(plan.description))

    click.echo(click.style("Command: ", bold=True, fg="magenta") + click.style(plan.full_command, bold=True))
    click.echo(click.style("Success Codes: ", bold=True, fg="magenta") + click.style(plan.success_codes, bold=True))

    if plan.inputs:
        click.echo(click.style("Inputs: ", bold=True, fg="magenta"))
        for run_input in plan.inputs:
            click.echo(click.style(f"\t- {run_input.name}:", bold=True))

            if run_input.description:
                click.echo(click.style(f"\t\t{run_input.description}"))

            click.echo(
                click.style("\t\tDefault Value: ", bold=True, fg="magenta")
                + click.style(run_input.default_value, bold=True)
            )

            if run_input.position:
                click.echo(
                    click.style("\t\tPosition: ", bold=True, fg="magenta") + click.style(run_input.position, bold=True)
                )

            if run_input.prefix:
                click.echo(
                    click.style("\t\tPrefix: ", bold=True, fg="magenta") + click.style(run_input.prefix, bold=True)
                )

    if plan.outputs:
        click.echo(click.style("Outputs: ", bold=True, fg="magenta"))
        for run_output in plan.outputs:
            click.echo(click.style(f"\t- {run_output.name}:", bold=True))

            if run_output.description:
                click.echo(click.style(f"\t\t{run_output.description}"))

            click.echo(
                click.style("\t\tDefault Value: ", bold=True, fg="magenta")
                + click.style(run_output.default_value, bold=True)
            )

            if run_output.position:
                click.echo(
                    click.style("\t\tPosition: ", bold=True, fg="magenta") + click.style(run_output.position, bold=True)
                )

            if run_output.prefix:
                click.echo(
                    click.style("\t\tPrefix: ", bold=True, fg="magenta") + click.style(run_output.prefix, bold=True)
                )

    if plan.parameters:
        click.echo(click.style("Parameters: ", bold=True, fg="magenta"))
        for run_parameter in plan.parameters:
            click.echo(click.style(f"\t- {run_parameter.name}:", bold=True))

            if run_parameter.description:
                click.echo(click.style(f"\t\t{run_parameter.description}"))

            click.echo(
                click.style("\t\tDefault Value: ", bold=True, fg="magenta")
                + click.style(run_parameter.default_value, bold=True)
            )

            if run_parameter.position:
                click.echo(
                    click.style("\t\tPosition: ", bold=True, fg="magenta")
                    + click.style(run_parameter.position, bold=True)
                )

            if run_parameter.prefix:
                click.echo(
                    click.style("\t\tPrefix: ", bold=True, fg="magenta") + click.style(run_parameter.prefix, bold=True)
                )


def _print_composite_plan(composite_plan: CompositePlanViewModel):
    """Print a CompositePlan to stdout."""
    click.echo(click.style("Id: ", bold=True, fg="magenta") + click.style(composite_plan.id, bold=True))
    click.echo(click.style("Name: ", bold=True, fg="magenta") + click.style(composite_plan.name, bold=True))

    if composite_plan.description:
        Console().print(Markdown(composite_plan.description))

    click.echo(click.style("Steps: ", bold=True, fg="magenta"))
    for step in composite_plan.steps:
        click.echo(click.style(f"\t- {step.name}:", bold=True))
        click.echo(click.style("\t\tId: ", bold=True, fg="magenta") + click.style(f"{step.id}", bold=True))

    if composite_plan.mappings:
        click.echo(click.style("Mappings: ", bold=True, fg="magenta"))
        for mapping in composite_plan.mappings:
            click.echo(click.style(f"\t- {mapping.name}:", bold=True))

            if mapping.description:
                click.echo(click.style(f"\t\t{mapping.description}"))

            click.echo(
                click.style("\t\tDefault Value: ", bold=True, fg="magenta")
                + click.style(mapping.default_value, bold=True)
            )
            click.echo(click.style("\tMaps to: ", bold=True, fg="magenta"))
            for maps_to in mapping.maps_to:
                click.style(maps_to, bold=True)

    if composite_plan.links:
        click.echo(click.style("Links: ", bold=True, fg="magenta"))
        for link in composite_plan.links:
            click.echo(click.style("\t- From: ", bold=True, fg="magenta") + click.style(link.source, bold=True))
            click.echo(click.style("\t\t To: ", bold=True, fg="magenta"))
            for sink in link.sinks:
                click.echo(click.style(f"\t\t- {sink}", bold=True))


@click.group()
def workflow():
    """Workflow commands."""
    pass


@workflow.command("ls")
@click.option("--format", type=click.Choice(WORKFLOW_FORMATS), default="tabular", help="Choose an output format.")
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="id,name",
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(WORKFLOW_COLUMNS.keys())),
    show_default=True,
)
def list_workflows(format, columns):
    """List or manage workflows with subcommands."""
    result = list_workflows_command().build().execute(format=format, columns=columns)
    click.echo(result.output)


@workflow.command()
@click.argument("name_or_id", metavar="<name_or_id>")
def show(name_or_id):
    """Show details for workflow <name_or_id>."""
    plan = show_workflow_command().build().execute(name_or_id=name_or_id).output

    if plan:
        if isinstance(plan, PlanViewModel):
            _print_plan(plan)
        else:
            _print_composite_plan(plan)
    else:
        click.secho(ERROR + f"Workflow '{name_or_id}' not found.")


@workflow.command()
@click.argument("name", metavar="<name>")
@click.option("--force", is_flag=True, help="Override the existence check.")
def remove(name, force):
    """Remove a workflow named <name>."""
    remove_workflow_command().build().execute(name=name, force=force)


@workflow.command()
@click.option("-d", "--description", help="Workflow step's description.")
@click.option("mappings", "-m", "--map", multiple=True, help="Mapping for a workflow parameter.")
@click.option("defaults", "-s", "--set", multiple=True, help="Default value for a workflow parameter.")
@click.option("links", "-l", "--link", multiple=True, help="Link source and sink parameters to connect steps.")
@click.option("-p", "--describe-param", multiple=True, help="Add description for a workflow parameter.")
@click.option("--map-inputs", is_flag=True, help="Exposes all child inputs as inputs on the CompositePlan.")
@click.option("--map-outputs", is_flag=True, help="Exposes all child outputs as outputs on the CompositePlan.")
@click.option("--map-params", is_flag=True, help="Exposes all child parameters as parameters on the CompositePlan.")
@click.option("--map-all", is_flag=True, help="Combination of --map-inputs, --map-outputs, --map-params.")
@click.option("--link-all", is_flag=True, help="Automatically link steps based on default values.")
@click.option("--keyword", multiple=True, help="List of keywords for the workflow.")
@click.argument("name", required=True)
@click.argument("steps", nargs=-1, required=True, type=click.UNPROCESSED)
def compose(
    description,
    mappings,
    defaults,
    links,
    describe_param,
    map_inputs,
    map_outputs,
    map_params,
    map_all,
    link_all,
    keyword,
    name,
    steps,
):
    """Create a composite workflow consisting of multiple steps."""

    if map_all:
        map_inputs = map_outputs = map_params = True

    result = (
        compose_workflow_command()
        .build()
        .execute(
            name=name,
            description=description,
            mappings=mappings,
            defaults=defaults,
            links=links,
            param_descriptions=describe_param,
            map_inputs=map_inputs,
            map_outputs=map_outputs,
            map_params=map_params,
            link_all=link_all,
            keywords=keyword,
            steps=steps,
        )
    )

    if not result.error:
        _print_composite_plan(result.output)


@workflow.command()
@click.argument("workflow_name", metavar="<name or uuid>")
@click.option("--name", metavar="<new name>", help="New name of the workflow")
@click.option("--description", metavar="<new desc>", help="New description of the workflow")
@click.option(
    "--set",
    "set_params",
    multiple=True,
    metavar="<parameter>=<value>",
    help="Set default <value> for a <parameter>/add new parameter",
)
@click.option(
    "--map",
    "map_params",
    multiple=True,
    metavar="<parameter>=<parameter or expression>",
    help="New mapping on the workflow",
)
@click.option(
    "--rename-param",
    "rename_params",
    multiple=True,
    metavar='<parameter>="name"',
    help="New name for parameter",
)
@click.option(
    "--describe-param",
    "describe_params",
    multiple=True,
    metavar='<parameter>="description"',
    help="New description of the workflow",
)
def edit(workflow_name, name, description, set_params, map_params, rename_params, describe_params):
    """Edit workflow details."""
    result = (
        edit_workflow_command()
        .build()
        .execute(
            name=workflow_name,
            new_name=name,
            description=description,
            set_params=set_params,
            map_params=map_params,
            rename_params=rename_params,
            describe_params=describe_params,
        )
    )
    if not result.error:
        plan = result.output
        if isinstance(plan, PlanViewModel):
            _print_plan(plan)
        else:
            _print_composite_plan(plan)


@workflow.command()
@click.argument("workflow_name", metavar="<name or uuid>")
@click.option(
    "--format",
    default="cwl",
    type=click.Choice(supported_formats(), case_sensitive=False),
    show_default=True,
    help="Workflow language format.",
)
@click.option(
    "-o",
    "--output",
    metavar="<path>",
    type=click.Path(exists=False),
    default=None,
    help="Save to <path> instead of printing to terminal",
)
@click.option(
    "--values",
    metavar="<file>",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="YAML file containing parameter mappings to be used.",
)
def export(workflow_name, format, output, values):
    """Export workflow."""
    communicator = ClickCallback()

    result = (
        export_workflow_command()
        .with_communicator(communicator)
        .build()
        .execute(name_or_id=workflow_name, format=format, output=output, values=values)
    )

    if not output:
        click.echo(result.output)


@workflow.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.pass_context
def inputs(ctx, paths):
    """Show all inputs used by workflows.

    <PATHS>    Limit results to these paths.
    """
    result = workflow_inputs_command().build().execute(paths=paths)

    input_paths = result.output

    click.echo("\n".join(input_paths))

    if paths:
        if not input_paths or any(
            p not in input_paths and all(Path(o) not in Path(p).parents for o in input_paths) for p in paths
        ):
            ctx.exit(1)


@workflow.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.pass_context
def outputs(ctx, paths):
    """Show all outputs generated by workflows.

    <PATHS>    Limit results to these paths.
    """
    result = workflow_outputs_command().build().execute(paths=paths)

    output_paths = result.output

    click.echo("\n".join(output_paths))

    if paths:
        if not output_paths or any(
            p not in output_paths and all(Path(o) not in Path(p).parents for o in output_paths) for p in paths
        ):
            ctx.exit(1)


@workflow.command()
@click.option(
    "provider",
    "-p",
    "--provider",
    default="cwltool",
    show_default=True,
    type=click.Choice(available_workflow_providers(), case_sensitive=False),
    help="The workflow engine to use.",
)
@click.option("config", "-c", "--config", metavar="<config file>", help="YAML file containing config for the provider.")
@click.option(
    "set_params",
    "-s",
    "--set",
    multiple=True,
    metavar="<parameter>=<value>",
    help="Set <value> for a <parameter> to be used in execution.",
)
@click.option(
    "--values",
    metavar="<file>",
    type=click.Path(exists=True, dir_okay=False),
    help="YAML file containing parameter mappings to be used.",
)
@click.argument("name_or_id", required=True)
def execute(
    provider,
    config,
    set_params,
    values,
    name_or_id,
):
    """Execute a given workflow."""
    communicator = ClickCallback()

    result = (
        execute_workflow_command()
        .with_communicator(communicator)
        .build()
        .execute(
            name_or_id=name_or_id,
            provider=provider,
            config=config,
            values=values,
            set_params=set_params,
        )
    )

    if result.output:
        click.echo(
            "Unchanged files:\n\n\t{0}".format("\n\t".join(click.style(path, fg="yellow") for path in result.output))
        )



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers utils for interacting with remote source code management tools."""
import re
from functools import reduce
from pathlib import Path

from renku.core.errors import ParameterError


def is_ascii(data):
    """Check if provided string contains only ascii characters."""
    return len(data) == len(data.encode())


def normalize_to_ascii(input_string, sep="-"):
    """Adjust chars to make the input compatible as scm source."""
    replace_all = [sep, "_", "."]
    for replacement in replace_all:
        input_string = input_string.replace(replacement, " ")

    return (
        sep.join(
            [
                component
                for component in re.sub(r"[^a-zA-Z0-9_.-]+", " ", input_string).split(" ")
                if component and is_ascii(component)
            ]
        )
        .lower()
        .strip(sep)
    )


def git_unicode_unescape(s: str, encoding: str = "utf-8") -> str:
    """Undoes git/gitpython unicode encoding."""
    if s.startswith('"'):
        return s.strip('"').encode("latin1").decode("unicode-escape").encode("latin1").decode(encoding)
    return s


def shorten_message(message: str, line_length: int = 100, body_length: int = 65000):
    """Wraps and shortens a commit message.

    :param message: message to adjust.
    :param line_length: maximum line length before wrapping. 0 for infinite.
    :param body_length: maximum body length before cut. 0 for infinite.
    :return: message wrapped and trimmed.
    """

    if line_length < 0:
        raise ParameterError("the length can't be negative.", "line_length")

    if body_length < 0:
        raise ParameterError("the length can't be negative.", "body_length")

    if body_length and len(message) > body_length:
        message = message[: body_length - 3] + "..."

    if line_length == 0 or len(message) <= line_length:
        return message

    lines = message.split(" ")
    lines = [
        line
        if len(line) < line_length
        else "\n\t".join(line[o : o + line_length] for o in range(0, len(line), line_length))
        for line in lines
    ]

    # NOTE: tries to preserve message spacing.
    wrapped_message = reduce(
        lambda c, x: (f"{c[0]} {x}", c[1] + len(x) + 1)
        if c[1] + len(x) <= line_length
        else (f"{c[0]}\n\t" + x, len(x)),
        lines,
        ("", 0),
    )[0]
    return wrapped_message[1:]


def safe_path(filepath):
    """Check if the path should be used in output."""
    if isinstance(filepath, Path):
        filepath = str(filepath)

    # Should not be in ignore paths.
    if filepath in {".gitignore", ".gitattributes"}:
        return False

    # Ignore everything in .renku ...
    if filepath.startswith(".renku"):
        return False

    return True



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku fixtures for repository management."""
import contextlib
import os
import secrets
import shutil

import pytest

from tests.utils import format_result_exception


@contextlib.contextmanager
def _isolated_filesystem(tmpdir, name=None, delete=True):
    """Click CliRunner ``isolated_filesystem`` but xdist compatible."""
    from renku.core.utils.contexts import chdir

    if not name:
        name = secrets.token_hex(8)
    t = tmpdir.mkdir(name)

    with chdir(t):
        try:
            yield t
        finally:
            if delete:
                try:
                    shutil.rmtree(t)
                except OSError:  # noqa: B014
                    pass


@pytest.fixture()
def renku_path(tmpdir):
    """Temporary instance path."""
    path = str(tmpdir.mkdir("renku"))
    yield path
    shutil.rmtree(path)


@pytest.fixture()
def instance_path(renku_path, monkeypatch):
    """Temporary instance path."""
    with monkeypatch.context() as m:
        m.chdir(renku_path)
        yield renku_path


@pytest.fixture
def repository(tmpdir):
    """Yield a Renku repository."""
    from click.testing import CliRunner
    from git.config import GitConfigParser, get_config_path

    from renku.cli import cli

    runner = CliRunner()
    with _isolated_filesystem(tmpdir, delete=True) as project_path:
        home = tmpdir.mkdir("user_home")
        old_home = os.environ.get("HOME", "")
        old_xdg_home = os.environ.get("XDG_CONFIG_HOME", "")

        try:
            # NOTE: fake user home directory
            os.environ["HOME"] = str(home)
            os.environ["XDG_CONFIG_HOME"] = str(home)
            with GitConfigParser(get_config_path("global"), read_only=False) as global_config:
                global_config.set_value("user", "name", "Renku @ SDSC")
                global_config.set_value("user", "email", "renku@datascience.ch")

            result = runner.invoke(cli, ["init", ".", "--template-id", "python-minimal"], "\n", catch_exceptions=False)
            assert 0 == result.exit_code, format_result_exception(result)

            yield os.path.realpath(project_path)
        finally:
            os.environ["HOME"] = old_home
            os.environ["XDG_CONFIG_HOME"] = old_xdg_home
            try:
                shutil.rmtree(home)
            except OSError:  # noqa: B014
                pass


@pytest.fixture
def project(repository):
    """Create a test project."""
    from click.testing import CliRunner
    from git import Repo

    from renku.cli import cli
    from renku.core.utils.contexts import chdir

    runner = CliRunner()

    repo = Repo(repository, search_parent_directories=True)
    commit = repo.head.commit

    with chdir(repository):
        yield repository

        os.chdir(repository)
        repo.head.reset(commit, index=True, working_tree=True)
        # INFO: remove any extra non-tracked files (.pyc, etc)
        repo.git.clean("-xdff")

        assert 0 == runner.invoke(cli, ["githooks", "install", "--force"]).exit_code


@pytest.fixture
def client(project, global_config_dir):
    """Return a Renku repository."""
    from renku.core.management import LocalClient
    from renku.core.models.enums import ConfigFilter

    original_get_value = LocalClient.get_value

    def mocked_get_value(self, section, key, config_filter=ConfigFilter.ALL):
        """We don't want lfs warnings in tests."""
        if key == "show_lfs_message":
            return "False"
        return original_get_value(self, section, key, config_filter=config_filter)

    LocalClient.get_value = mocked_get_value

    yield LocalClient(path=project)

    LocalClient.get_value = original_get_value


@pytest.fixture
def client_injection_bindings():
    """Return bindings needed for client dependency injection."""

    def _create_client_bindings(client):
        from renku.core.management.command_builder.client_dispatcher import ClientDispatcher
        from renku.core.management.interface.client_dispatcher import IClientDispatcher

        client_dispatcher = ClientDispatcher()
        client_dispatcher.push_created_client_to_stack(client)
        return {"bindings": {"LocalClient": client, IClientDispatcher: client_dispatcher}, "constructor_bindings": {}}

    return _create_client_bindings


@pytest.fixture
def injection_binder(request):
    """Return a binder that can work with bindings."""

    def _binder(bindings):
        from renku.core.management.command_builder.command import inject, remove_injector

        def _bind(binder):
            for key, value in bindings["bindings"].items():
                binder.bind(key, value)
            for key, value in bindings["constructor_bindings"].items():
                binder.bind_to_constructor(key, value)

            return binder

        inject.configure(_bind, bind_in_runtime=False)
        request.addfinalizer(lambda: remove_injector())
        return

    return _binder


@pytest.fixture
def injected_client(client, client_injection_bindings, injection_binder):
    """Inject a client."""
    injection_binder(client_injection_bindings(client))



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku save commands."""

from functools import reduce
from uuid import uuid4

import git

from renku.core import errors
from renku.core.management.command_builder import inject
from renku.core.management.command_builder.command import Command
from renku.core.management.interface.client_dispatcher import IClientDispatcher
from renku.core.utils import communication
from renku.core.utils.git import add_to_git
from renku.core.utils.scm import git_unicode_unescape


@inject.autoparams()
def _save_and_push(client_dispatcher: IClientDispatcher, message=None, remote=None, paths=None):
    """Save and push local changes."""
    client = client_dispatcher.current_client

    client.setup_credential_helper()
    if not paths:
        paths = client.dirty_paths
    else:
        staged = client.repo.index.diff("HEAD")
        if staged:
            staged = {git_unicode_unescape(p.a_path) for p in staged}
            not_passed = staged - set(paths)

            if not_passed:
                raise errors.RenkuSaveError(
                    "These files are in the git staging area, but weren't passed to renku save. Unstage them or pass"
                    + " them explicitly: \n"
                    + "\n".join(not_passed)
                )

    if paths:
        client.track_paths_in_storage(*paths)

    return repo_sync(client.repo, message, remote, paths)


def save_and_push_command():
    """Command to save and push."""
    return Command().command(_save_and_push)


def repo_sync(repo, message=None, remote=None, paths=None):
    """Commit and push paths."""
    origin = None
    saved_paths = []

    # get branch that's pushed
    if repo.active_branch.tracking_branch():
        ref = repo.active_branch.tracking_branch().name
        pushed_branch = ref.split("/")[-1]
    else:
        pushed_branch = repo.active_branch.name

    if remote:
        # get/setup supplied remote for pushing
        if repo.remotes:
            existing = next((r for r in repo.remotes if r.url == remote), None)
            if not existing:
                existing = next((r for r in repo.remotes if r.name == remote), None)
            origin = next((r for r in repo.remotes if r.name == "origin"), None)
            if existing:
                origin = existing
            elif origin:
                pushed_branch = uuid4().hex
                origin = repo.create_remote(pushed_branch, remote)
        if not origin:
            origin = repo.create_remote("origin", remote)
    elif not repo.active_branch.tracking_branch():
        # No remote set on branch, push to available remote if only a single
        # one is available
        if len(repo.remotes) == 1:
            origin = repo.remotes[0]
        else:
            raise errors.ConfigurationError("No remote has been set up for the current branch")
    else:
        # get remote that's set up to track the local branch
        origin = repo.remotes[repo.active_branch.tracking_branch().remote_name]

    if paths:
        # commit uncommitted changes
        try:
            staged_files = (
                {git_unicode_unescape(d.a_path) for d in repo.index.diff("HEAD")} if repo.head.is_valid() else set()
            )

            path_to_save = set(paths) - staged_files

            if path_to_save:
                add_to_git(repo.git, *path_to_save)

            saved_paths = [d.b_path for d in repo.index.diff("HEAD")]

            if not message:
                # Show saved files in message
                max_len = 100
                message = "Saved changes to: "
                paths_with_lens = reduce(
                    lambda c, x: c + [(x, c[-1][1] + len(x))], saved_paths, [(None, len(message))]
                )[1:]
                # limit first line to max_len characters
                message += " ".join(p if l < max_len else "\n\t" + p for p, l in paths_with_lens)

            repo.index.commit(message)
        except git.exc.GitCommandError as e:
            raise errors.GitError("Cannot commit changes") from e

    try:
        # NOTE: Push local changes to remote branch.
        merge_conflict = False
        if origin.refs and repo.active_branch.tracking_branch() and repo.active_branch.tracking_branch() in origin.refs:
            origin.fetch()
            try:
                origin.pull(repo.active_branch)
            except git.exc.GitCommandError:
                # NOTE: Couldn't pull, probably due to conflicts, try a merge.
                # NOTE: the error sadly doesn't tell any details.
                unmerged_blobs = repo.index.unmerged_blobs().values()
                conflicts = (stage != 0 for blobs in unmerged_blobs for stage, _ in blobs)
                if any(conflicts):
                    merge_conflict = True

                    if communication.confirm(
                        "There were conflicts when updating the local data with remote changes,"
                        " do you want to resolve them (if not, a new remote branch will be created)?",
                        warning=True,
                    ):
                        repo.git.mergetool("-g")
                        repo.git.commit("--no-edit")
                        merge_conflict = False
                    else:
                        repo.head.reset(index=True, working_tree=True)
                else:
                    raise

        result = None
        failed_push = None

        if not merge_conflict:
            result = origin.push(repo.active_branch)
            failed_push = [
                push_info
                for push_info in result
                if push_info.flags & git.PushInfo.ERROR
                or push_info.flags & git.PushInfo.REJECTED
                or push_info.flags & git.PushInfo.REMOTE_REJECTED
                or push_info.flags & git.PushInfo.REMOTE_FAILURE
            ]

        if merge_conflict or (result and "[remote rejected] (pre-receive hook declined)" in result[0].summary):
            # NOTE: Push to new remote branch if original one is protected and reset the cache.
            old_pushed_branch = pushed_branch
            old_active_branch = repo.active_branch
            pushed_branch = uuid4().hex
            try:
                repo.create_head(pushed_branch)
                result = repo.remote().push(pushed_branch)
                failed_push = [
                    push_info
                    for push_info in result
                    if push_info.flags & git.PushInfo.ERROR
                    or push_info.flags & git.PushInfo.REJECTED
                    or push_info.flags & git.PushInfo.REMOTE_REJECTED
                    or push_info.flags & git.PushInfo.REMOTE_FAILURE
                ]
            finally:
                # Reset cache
                repo.git.checkout(old_active_branch)
                ref = f"{origin}/{old_pushed_branch}"
                repo.head.reset(commit=ref, index=True, working_tree=True)

        if result and failed_push:
            # NOTE: Couldn't push for some reason
            msg = "\n".join(info.summary for info in failed_push)
            raise errors.GitError(f"Couldn't push changes. Reason:\n{msg}")

    except git.exc.GitCommandError as e:
        raise errors.GitError("Cannot push changes") from e

    return saved_paths, pushed_branch



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service graph jobs tests."""
import json

import pytest


@pytest.mark.service
@pytest.mark.jobs
@pytest.mark.integration
@pytest.mark.parametrize("revision", [None, "HEAD", "HEAD^^", "HEAD^^..HEAD"])
def test_graph_export_job(svc_client_cache, it_remote_repo_url, revision):
    """Test graph export job."""
    svc_client, headers, _ = svc_client_cache

    payload = {
        "git_url": it_remote_repo_url,
        "revision": revision,
        "callback_url": "https://webhook.site",
        "migrate_project": True,
    }

    response = svc_client.get("/graph.export", data=json.dumps(payload), headers=headers)
    assert response
    assert {"graph"} == set(response.json["result"].keys())



# -*- coding: utf-8 -*-
#
# Copyright 2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers functions for metadata management/parsing."""

from collections.abc import Iterable
from typing import List, Set, Tuple, Union

from git import Repo

from renku.core import errors
from renku.core.models.entity import Entity
from renku.core.models.provenance.activity import Activity
from renku.core.models.provenance.agent import Person
from renku.core.utils.git import get_object_hash


def construct_creators(creators: List[Union[dict, str]], ignore_email=False):
    """Parse input and return a list of Person."""
    creators = creators or ()

    if not isinstance(creators, Iterable) or isinstance(creators, str):
        raise errors.ParameterError("Invalid creators type")

    people = []
    no_email_warnings = []
    for creator in creators:
        person, no_email_warning = construct_creator(creator, ignore_email=ignore_email)

        people.append(person)

        if no_email_warning:
            no_email_warnings.append(no_email_warning)

    return people, no_email_warnings


def construct_creator(creator: Union[dict, str], ignore_email):
    """Parse input and return an instance of Person."""
    if not creator:
        return None, None

    if isinstance(creator, str):
        person = Person.from_string(creator)
    elif isinstance(creator, dict):
        person = Person.from_dict(creator)
    else:
        raise errors.ParameterError("Invalid creator type")

    message = 'A valid format is "Name <email> [affiliation]"'

    if not person.name:  # pragma: no cover
        raise errors.ParameterError(f'Name is invalid: "{creator}".\n{message}')

    if not person.email:
        if not ignore_email:  # pragma: no cover
            raise errors.ParameterError(f'Email is invalid: "{creator}".\n{message}')
        else:
            no_email_warning = creator
    else:
        no_email_warning = None

    return person, no_email_warning


def get_modified_activities(
    activities: List[Activity], repo: Repo
) -> Tuple[Set[Tuple[Activity, Entity]], Set[Tuple[Activity, Entity]]]:
    """Get lists of activities that have modified/deleted usage entities."""
    modified = set()
    deleted = set()

    for activity in activities:
        for usage in activity.usages:
            entity = usage.entity
            current_checksum = get_object_hash(repo=repo, path=entity.path)
            if current_checksum is None:
                deleted.add((activity, entity))
            elif current_checksum != entity.checksum:
                modified.add((activity, entity))

    return modified, deleted


def add_activity_if_recent(activity: Activity, activities: Set[Activity]):
    """Add ``activity`` to ``activities`` if it's not in the set or is the latest executed instance."""
    if activity in activities:
        return

    for existing_activity in activities:
        if activity.has_identical_inputs_and_outputs_as(existing_activity):
            if activity.ended_at_time > existing_activity.ended_at_time:  # activity is newer
                activities.remove(existing_activity)
                activities.add(activity)
            return

    # NOTE: No similar activity was found
    activities.add(activity)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent run templates."""

import copy
import itertools
import re
from abc import ABC
from datetime import datetime
from typing import Any, List, Tuple
from uuid import uuid4

import marshmallow
from werkzeug.utils import secure_filename

from renku.core import errors
from renku.core.metadata.database import Persistent
from renku.core.models.calamus import JsonLDSchema, Nested, fields, prov, renku, schema
from renku.core.models.workflow.parameter import (
    CommandInput,
    CommandInputSchema,
    CommandOutput,
    CommandOutputSchema,
    CommandParameter,
    CommandParameterBase,
    CommandParameterSchema,
)

MAX_GENERATED_NAME_LENGTH = 25


class AbstractPlan(Persistent, ABC):
    """Abstract base class for all plans."""

    def __init__(
        self,
        *,
        description: str = None,
        id: str,
        invalidated_at: datetime = None,
        keywords: List[str] = None,
        name: str = None,
        project_id: str = None,
        derived_from: str = None,
    ):
        self.description: str = description
        self.id: str = id
        self.invalidated_at: datetime = invalidated_at
        self.keywords: List[str] = keywords or []
        self.name: str = name
        self.project_id: str = project_id
        self.derived_from: str = derived_from

        if not self.name:
            self.name = self._get_default_name()
        else:
            AbstractPlan.validate_name(name)

    @staticmethod
    def generate_id(uuid: str = None) -> str:
        """Generate an identifier for Plan."""
        uuid = uuid or uuid4().hex
        return f"/plans/{uuid}"

    def _get_default_name(self) -> str:
        name = "-".join(str(a) for a in self.to_argv())
        if not name:
            return uuid4().hex[:MAX_GENERATED_NAME_LENGTH]

        name = secure_filename(name)
        rand_length = 5
        return f"{name[:MAX_GENERATED_NAME_LENGTH - rand_length -1]}-{uuid4().hex[:rand_length]}"

    @staticmethod
    def validate_name(name: str):
        """Check a name for invalid characters."""
        if not re.match("[a-zA-Z0-9-_]+", name):
            raise errors.ParameterError(
                f"Name {name} contains illegal characters. Only characters, numbers, _ and - are allowed."
            )

    def assign_new_id(self) -> str:
        """Assign a new UUID.

        This is required only when there is another plan which is exactly the same except the parameters' list.
        """
        current_uuid = self._extract_uuid()
        new_uuid = uuid4().hex
        self.id = self.id.replace(current_uuid, new_uuid)

        return new_uuid

    def _extract_uuid(self) -> str:
        path_start = self.id.find("/plans/")
        return self.id[path_start + len("/plans/") :]

    def resolve_mapping_path(self, mapping_path: str) -> Tuple[CommandParameterBase, "Plan"]:
        """Resolve a mapping path to its reference parameter."""
        raise NotImplementedError()

    def resolve_direct_reference(self, reference: str) -> CommandParameterBase:
        """Resolve a direct parameter reference."""
        raise NotImplementedError()

    def find_parameter(self, parameter: CommandParameterBase) -> bool:
        """Find if a parameter exists on this plan."""
        raise NotImplementedError()

    def find_parameter_workflow(self, parameter: CommandParameterBase) -> "Plan":
        """Return the workflow a parameter belongs to."""
        raise NotImplementedError()


class Plan(AbstractPlan):
    """Represent a `renku run` execution template."""

    def __init__(
        self,
        *,
        parameters: List[CommandParameter] = None,
        command: str = None,
        description: str = None,
        id: str,
        inputs: List[CommandInput] = None,
        invalidated_at: datetime = None,
        keywords: List[str] = None,
        name: str = None,
        derived_from: str = None,
        project_id: str = None,
        outputs: List[CommandOutput] = None,
        success_codes: List[int] = None,
    ):
        self.command: str = command
        self.inputs: List[CommandInput] = inputs or []
        self.outputs: List[CommandOutput] = outputs or []
        self.parameters: List[CommandParameter] = parameters or []
        self.success_codes: List[int] = success_codes or []
        super().__init__(
            id=id,
            description=description,
            invalidated_at=invalidated_at,
            keywords=keywords,
            name=name,
            project_id=project_id,
            derived_from=derived_from,
        )

    def is_similar_to(self, other: "Plan") -> bool:
        """Return true if plan has the same inputs/outputs/arguments as another plan."""

        def get_input_patterns(plan: Plan):
            return {e.default_value for e in plan.inputs}

        def get_output_patterns(plan: Plan):
            return {e.default_value for e in plan.outputs}

        def get_parameters(plan: Plan):
            return {(a.position, a.prefix, a.default_value) for a in plan.parameters}

        # TODO: Check order of inputs/outputs/parameters as well after sorting by position
        return (
            self.command == other.command
            and set(self.success_codes) == set(other.success_codes)
            and get_input_patterns(self) == get_input_patterns(other)
            and get_output_patterns(self) == get_output_patterns(other)
            and get_parameters(self) == get_parameters(other)
        )

    def resolve_mapping_path(self, mapping_path: str) -> Tuple[CommandParameterBase, "Plan"]:
        """Resolve a mapping path to its reference parameter."""

        parts = mapping_path.split(".", maxsplit=1)

        if len(parts) > 1:
            raise errors.ParameterNotFoundError(mapping_path, self.name)

        return self.resolve_direct_reference(parts[0]), self

    def resolve_direct_reference(self, reference: str) -> CommandParameterBase:
        """Resolve a direct parameter reference."""
        try:
            if reference.startswith("@input"):
                return self.inputs[int(reference[6:]) - 1]
            elif reference.startswith("@output"):
                return self.outputs[int(reference[7:]) - 1]
            elif reference.startswith("@param"):
                return self.parameters[int(reference[6:]) - 1]
        except (ValueError, IndexError):
            raise errors.ParameterNotFoundError(reference, self.name)

        for parameter in self.inputs + self.outputs + self.parameters:
            if parameter.name == reference:
                return parameter

        raise errors.ParameterNotFoundError(reference, self.name)

    def find_parameter(self, parameter: CommandParameterBase) -> bool:
        """Find if a parameter exists on this plan."""
        return parameter in self.inputs + self.outputs + self.parameters

    def find_parameter_workflow(self, parameter: CommandParameterBase) -> "Plan":
        """Return the workflow a parameter belongs to."""
        if self.find_parameter(parameter):
            return self

    def assign_new_id(self):
        """Assign a new UUID.

        This is required only when there is another plan which is exactly the same except the parameters' list.
        """
        new_uuid = super().assign_new_id()
        current_uuid = self._extract_uuid()
        if hasattr(self, "parameters"):
            self.parameters = copy.deepcopy(self.parameters)

            for a in self.parameters:
                a.id = a.id.replace(current_uuid, new_uuid)

    def derive(self) -> "Plan":
        """Create a new ``Plan`` that is derived from self."""
        derived = Plan(
            parameters=self.parameters.copy(),
            command=self.command,
            description=self.description,
            id=self.id,
            inputs=self.inputs.copy(),
            invalidated_at=self.invalidated_at,
            keywords=self.keywords.copy(),
            name=self.name,
            derived_from=self.id,
            outputs=self.outputs.copy(),
            success_codes=self.success_codes.copy(),
        )
        derived.assign_new_id()
        return derived

    @property
    def keywords_csv(self):
        """Comma-separated list of keywords associated with workflow."""
        return ", ".join(self.keywords)

    def to_argv(self) -> List[Any]:
        """Convert a Plan into argv list."""
        arguments = itertools.chain(self.inputs, self.outputs, self.parameters)
        arguments = filter(lambda a: a.position is not None and not getattr(a, "mapped_to", None), arguments)
        arguments = sorted(arguments, key=lambda a: a.position)

        argv = self.command.split(" ") if self.command else []
        argv.extend(e for a in arguments for e in a.to_argv())

        return argv

    def set_parameters_from_strings(self, params_strings: List[str]) -> None:
        """Set parameters by parsing parameters strings."""
        for param_string in params_strings:
            name, value = param_string.split("=", maxsplit=1)
            for param in self.inputs + self.outputs + self.parameters:
                if param.name == name:
                    param.default_value = value
                    break
            else:
                self.parameters.append(
                    CommandParameter(default_value=value, id=CommandParameter.generate_id(plan_id=self.id), name=name)
                )

    def copy(self):
        """Create a copy of this plan.

        Required where a plan is used several times in a workflow but we need to set different values on them.
        """
        return copy.deepcopy(self)


class PlanSchema(JsonLDSchema):
    """Plan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork]
        model = Plan
        unknown = marshmallow.EXCLUDE

    command = fields.String(renku.command, missing=None)
    description = fields.String(schema.description, missing=None)
    id = fields.Id()
    inputs = Nested(renku.hasInputs, CommandInputSchema, many=True, missing=None)
    invalidated_at = fields.DateTime(prov.invalidatedAtTime, add_value_types=True)
    keywords = fields.List(schema.keywords, fields.String(), missing=None)
    name = fields.String(schema.name, missing=None)
    derived_from = fields.String(prov.wasDerivedFrom, missing=None)
    project_id = fields.IRI(renku.hasPlan, reverse=True)
    outputs = Nested(renku.hasOutputs, CommandOutputSchema, many=True, missing=None)
    parameters = Nested(renku.hasArguments, CommandParameterSchema, many=True, missing=None)
    success_codes = fields.List(renku.successCodes, fields.Integer(), missing=[0])


class PlanDetailsJson(marshmallow.Schema):
    """Serialize a plan to a response object."""

    name = marshmallow.fields.String(required=True)
    derived_from = marshmallow.fields.String()
    title = marshmallow.fields.String()
    description = marshmallow.fields.String()
    keywords = marshmallow.fields.List(marshmallow.fields.String())
    id = marshmallow.fields.String()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Activity."""

from datetime import datetime
from unittest.mock import MagicMock
from uuid import uuid4

from renku.core.models.entity import Entity
from renku.core.models.provenance.activity import Activity
from renku.core.models.provenance.agent import Person
from renku.core.models.workflow.parameter import CommandInput, CommandOutput, CommandParameter
from renku.core.models.workflow.plan import Plan


def test_activity_parameter_values(mocker):
    """Test parameter values are correctly set on an activity."""

    def mock_entity_from_revision(client, path, revision=None):
        return Entity(checksum="abcdefg", id=uuid4().hex, path=path)

    def mock_person_from_client(client):
        return Person(id=uuid4().hex, name="John Doe", email="john@doe.com")

    mocker.patch("renku.core.models.provenance.activity.Entity.from_revision", mock_entity_from_revision)
    mocker.patch("renku.core.models.provenance.activity.Person.from_client", mock_person_from_client)
    commit = MagicMock()
    commit.hexsha.return_value = uuid4().hex
    commit.committer.email.return_value = "john@doe.com"
    commit.committer.name.return_value = "John Doe"

    client_dispatcher = MagicMock()
    project_gateway = MagicMock()
    project_gateway.get_project.return_value.id.return_value = "some_project"

    i1 = CommandInput(id="i1", default_value="i1_path")
    i1_copy = CommandInput(id="i2", default_value="i1_path")
    i3 = CommandInput(id="i3", default_value="i3_path")
    i4 = CommandInput(id="i4", default_value="i4_path")
    i4.actual_value = "other_i4_path"

    o1 = CommandOutput(id="o1", default_value="o1_path")
    o1_copy = CommandOutput(id="o2", default_value="o1_path")
    o3 = CommandOutput(id="o3", default_value="o3_path")
    o4 = CommandOutput(id="o4", default_value="o4_path")
    o4.actual_value = "other_o4_path"

    p1 = CommandParameter(id="p1", default_value=1.1)
    p1_copy = CommandParameter(id="p2", default_value=1.1)
    p3 = CommandParameter(id="p3", default_value=3)
    p4 = CommandParameter(id="p4", default_value="4")
    p4.actual_value = "5"

    plan = Plan(
        id="test", inputs=[i1, i1_copy, i3, i4], outputs=[o1, o1_copy, o3, o4], parameters=[p1, p1_copy, p3, p4]
    )

    activity = Activity.from_plan(
        plan,
        client_dispatcher=client_dispatcher,
        project_gateway=project_gateway,
        started_at_time=datetime.utcnow(),
        ended_at_time=datetime.utcnow(),
        annotations=[],
        commit=commit,
    )

    assert len(activity.generations) == 3
    assert len(activity.usages) == 3

    assert len(activity.parameters) == 12

    pi1 = next(p for p in activity.parameters if i1.id == p.parameter_id)
    assert pi1.value == i1.default_value

    pi2 = next(p for p in activity.parameters if i1_copy.id == p.parameter_id)
    assert pi2.value == i1_copy.default_value

    pi3 = next(p for p in activity.parameters if i3.id == p.parameter_id)
    assert pi3.value == i3.default_value

    pi4 = next(p for p in activity.parameters if i4.id == p.parameter_id)
    assert pi4.value != i4.default_value
    assert pi4.value == i4.actual_value

    po1 = next(p for p in activity.parameters if o1.id == p.parameter_id)
    assert po1.value == o1.default_value

    po2 = next(p for p in activity.parameters if o1_copy.id == p.parameter_id)
    assert po2.value == o1_copy.default_value

    po3 = next(p for p in activity.parameters if o3.id == p.parameter_id)
    assert po3.value == o3.default_value

    po4 = next(p for p in activity.parameters if o4.id == p.parameter_id)
    assert po4.value != o4.default_value
    assert po4.value == o4.actual_value

    pp1 = next(p for p in activity.parameters if p1.id == p.parameter_id)
    assert pp1.value == p1.default_value

    pp2 = next(p for p in activity.parameters if p1_copy.id == p.parameter_id)
    assert pp2.value == p1_copy.default_value

    pp3 = next(p for p in activity.parameters if p3.id == p.parameter_id)
    assert pp3.value == p3.default_value

    pp4 = next(p for p in activity.parameters if p4.id == p.parameter_id)
    assert pp4.value != p4.default_value
    assert pp4.value == p4.actual_value

    # test getting applied plan
    pi1.value = "pi1"
    pi4.value = "pi4"
    po1.value = "po1"
    po4.value = "po4"
    pp1.value = "pp1"
    pp4.value = "pp4"

    applied_plan = activity.plan_with_values

    assert applied_plan.inputs[0].actual_value == pi1.value
    assert applied_plan.inputs[3].actual_value == pi4.value
    assert applied_plan.outputs[0].actual_value == po1.value
    assert applied_plan.outputs[3].actual_value == po4.value
    assert applied_plan.parameters[0].actual_value == pp1.value
    assert applied_plan.parameters[3].actual_value == pp4.value



# -*- coding: utf-8 -*-
#
# Copyright 2019-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project view tests."""
import json
import re

import pytest

from tests.utils import retry_failed


def assert_rpc_response(response, with_key="result"):
    """Check rpc result in response."""
    assert response and 200 == response.status_code

    response_text = re.sub(r"http\S+", "", json.dumps(response.json))
    assert with_key in response_text


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_edit_project_view(svc_client_with_repo):
    """Test editing project metadata."""
    svc_client, headers, project_id, _ = svc_client_with_repo

    edit_payload = {
        "project_id": project_id,
        "description": "my new title",
        "creator": {"name": "name123", "email": "name123@ethz.ch", "affiliation": "ethz"},
        "custom_metadata": {
            "@id": "http://example.com/metadata12",
            "@type": "https://schema.org/myType",
            "https://schema.org/property1": 1,
            "https://schema.org/property2": "test",
        },
    }
    response = svc_client.post("/project.edit", data=json.dumps(edit_payload), headers=headers)

    assert response
    assert_rpc_response(response)

    assert {"warning", "edited", "remote_branch"} == set(response.json["result"])
    assert {
        "description": "my new title",
        "creator": {"name": "name123", "email": "name123@ethz.ch", "affiliation": "ethz"},
        "custom_metadata": {
            "@id": "http://example.com/metadata12",
            "@type": "https://schema.org/myType",
            "https://schema.org/property1": 1,
            "https://schema.org/property2": "test",
        },
    } == response.json["result"]["edited"]


@pytest.mark.integration
@pytest.mark.service
@retry_failed
def test_remote_edit_view(svc_client, it_remote_repo_url, identity_headers):
    """Test creating a delayed edit."""
    response = svc_client.post(
        "/project.edit",
        data=json.dumps(dict(git_url=it_remote_repo_url, is_delayed=True)),
        headers=identity_headers,
    )

    assert 200 == response.status_code
    assert response.json["result"]["created_at"]
    assert response.json["result"]["job_id"]



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku core fixtures for workflow testing."""

import pytest

from renku.core.models.workflow.composite_plan import CompositePlan
from renku.core.models.workflow.parameter import CommandInput, CommandOutput, CommandParameter
from renku.core.models.workflow.plan import Plan


def _create_run(name: str) -> Plan:

    run_id = Plan.generate_id()
    input1 = CommandInput(
        id=CommandInput.generate_id(run_id, 1),
        position=1,
        default_value=1,
        name=f"{name}_input1",
    )
    input2 = CommandInput(
        id=CommandInput.generate_id(run_id, 2),
        position=2,
        default_value=2,
        name=f"{name}_input2",
    )

    output1 = CommandOutput(
        id=CommandOutput.generate_id(run_id, 3),
        position=3,
        default_value=3,
        name=f"{name}_output1",
    )
    output2 = CommandOutput(
        id=CommandOutput.generate_id(run_id, 4),
        position=4,
        default_value=4,
        name=f"{name}_output2",
    )

    param1 = CommandParameter(
        id=CommandParameter.generate_id(run_id, 5),
        position=5,
        default_value=5,
        name=f"{name}_param1",
    )
    param2 = CommandParameter(
        id=CommandParameter.generate_id(run_id, 6),
        position=6,
        default_value=6,
        name=f"{name}_param2",
    )

    return Plan(
        id=run_id,
        name=name,
        command="cat",
        inputs=[input1, input2],
        outputs=[output1, output2],
        parameters=[param1, param2],
    )


@pytest.fixture
def composite_plan():
    """Fixture for a basic CompositePlan."""
    run1 = _create_run("run1")
    run2 = _create_run("run2")

    grouped = CompositePlan(id=CompositePlan.generate_id(), plans=[run1, run2], name="grouped1")

    return grouped, run1, run2



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset tests."""
import copy
import datetime
import os
import shutil
import stat
from pathlib import Path

import pytest
from git import Repo

from renku.core import errors
from renku.core.commands.dataset import add_to_dataset, create_dataset, file_unlink, list_datasets, list_files
from renku.core.errors import ParameterError
from renku.core.management.datasets import DatasetsProvenance
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.models.dataset import Dataset
from renku.core.models.provenance.agent import Person
from renku.core.utils.contexts import chdir
from renku.core.utils.urls import get_slug
from tests.utils import assert_dataset_is_mutated, load_dataset, raises


@pytest.mark.serial
@pytest.mark.parametrize(
    "scheme, path, overwrite, error",
    [
        ("", "temp", False, None),
        ("file://", "temp", True, None),
        ("", "tempp", False, errors.ParameterError),
        ("http://", "example.com/file1", False, None),
        ("https://", "example.com/file1", True, None),
        ("bla://", "file", False, errors.UrlSchemeNotSupported),
    ],
)
def test_data_add(scheme, path, overwrite, error, client_with_injection, directory_tree, dataset_responses):
    """Test data import."""
    with raises(error):
        if path == "temp":
            path = str(directory_tree / "file1")

        with client_with_injection.with_dataset(name="dataset", create=True, commit_database=True) as d:
            d.creators = [Person(name="me", email="me@example.com", id="me_id")]
            client_with_injection.add_data_to_dataset(d, ["{}{}".format(scheme, path)], overwrite=overwrite)

        target_path = os.path.join(DATA_DIR, "dataset", "file1")

        with open(target_path) as f:
            assert f.read() == "123"

        assert d.find_file(target_path)

        # check that the imported file is read-only
        assert not os.access(target_path, stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        # check the linking
        if scheme in ("", "file://"):
            shutil.rmtree("./data/dataset")
            with client_with_injection.with_dataset(name="dataset") as d:
                d.creators = [Person(name="me", email="me@example.com", id="me_id")]
                client_with_injection.add_data_to_dataset(d, ["{}{}".format(scheme, path)], overwrite=True)
            assert os.path.exists(target_path)


def test_data_add_recursive(directory_tree, client_with_injection):
    """Test recursive data imports."""
    with client_with_injection.with_dataset(name="dataset", create=True) as dataset:
        dataset.creators = [Person(name="me", email="me@example.com", id="me_id")]
        client_with_injection.add_data_to_dataset(dataset, [str(directory_tree / "dir1")])

        assert os.path.basename(os.path.dirname(dataset.files[0].entity.path)) == "dir1"


@pytest.mark.parametrize(
    "creators",
    [
        [Person(name="me", email="me@example.com")],
        [Person.from_jsonld({"http://schema.org/name": "me", "http://schema.org/email": "me@example.com"})],
    ],
)
def test_creator_parse(creators):
    """Test that different options for specifying creators work."""
    dataset = Dataset(name="dataset", creators=creators)
    creator = Person(name="me", email="me@example.com")
    assert creator in dataset.creators

    # email check
    with pytest.raises(ValueError):
        Person(name="me", email="meexample.com")

    # creators must be a set or list of dicts or Person
    with pytest.raises(ValueError):
        Dataset(name="dataset", creators=["name"])


def test_creators_with_same_email(client_with_injection, load_dataset_with_injection):
    """Test creators with different names and same email address."""
    with client_with_injection.with_dataset(name="dataset", create=True, commit_database=True) as dataset:
        dataset.creators = [Person(name="me", email="me@example.com"), Person(name="me2", email="me@example.com")]
        DatasetsProvenance().add_or_update(dataset)

    dataset = load_dataset("dataset")

    assert 2 == len(dataset.creators)
    assert {c.name for c in dataset.creators} == {"me", "me2"}


def test_dataset_serialization(client_with_datasets, load_dataset_with_injection):
    """Test dataset (de)serialization."""
    dataset = load_dataset_with_injection("dataset-1", client_with_datasets)

    def read_value(key):
        return dataset_metadata.get(key)[0].get("@value")

    flattened_metadata = dataset.to_jsonld()
    dataset = Dataset.from_jsonld(flattened_metadata)

    # assert that all attributes found in metadata are set in the instance
    assert dataset.date_created
    assert dataset.creators
    assert dataset.identifier
    assert dataset.title

    dataset_metadata = [m for m in flattened_metadata if "@type" in m and "Dataset" in str(m["@type"])][0]

    # check values
    assert str(dataset.date_created.isoformat()) == read_value("http://schema.org/dateCreated")
    assert dataset.identifier == read_value("http://schema.org/identifier")
    assert dataset.title == read_value("http://schema.org/name")


def test_create_dataset_custom_message(project):
    """Test create dataset custom message."""
    create_dataset().with_commit_message("my dataset").with_database(write=True).build().execute(
        "ds1", title="", description="", creators=[]
    )

    last_commit = Repo(".").head.commit
    assert "my dataset" == last_commit.message


def test_list_datasets_default(project):
    """Test a default dataset listing."""
    create_dataset().with_commit_message("my dataset").with_database(write=True).build().execute(
        "ds1", title="", description="", creators=[]
    )

    datasets = list_datasets().with_database().build().execute().output

    assert isinstance(datasets, list)
    assert "ds1" in [dataset.title for dataset in datasets]


def test_list_files_default(project, tmpdir):
    """Test a default file listing."""
    create_dataset().with_commit_message("my dataset").with_database(write=True).build().execute(
        "ds1", title="", description="", creators=[]
    )
    data_file = tmpdir / Path("some-file")
    data_file.write_text("1,2,3", encoding="utf-8")

    add_to_dataset().build().execute([str(data_file)], "ds1")
    files = list_files().build().execute(datasets=["ds1"]).output

    assert isinstance(files, list)
    assert "some-file" in [Path(f.entity.path).name for f in files]


def test_unlink_default(directory_tree, client):
    """Test unlink default behaviour."""
    with chdir(client.path):
        create_dataset().with_database(write=True).build().execute("dataset")
        add_to_dataset().with_database(write=True).build().execute([str(directory_tree / "dir1")], "dataset")

    with pytest.raises(ParameterError):
        file_unlink().build().execute("dataset", (), ())


@pytest.mark.xfail
def test_mutate(client):
    """Test metadata change after dataset mutation."""
    dataset = Dataset(
        name="my-dataset",
        creators=[Person.from_string("John Doe <john.doe@mail.com>")],
        date_published=datetime.datetime.now(datetime.timezone.utc),
        same_as="http://some-url",
    )
    old_dataset = copy.deepcopy(dataset)

    dataset.mutate()

    mutator = Person.from_git(client.repo)
    assert_dataset_is_mutated(old=old_dataset, new=dataset, mutator=mutator)


@pytest.mark.xfail
def test_mutator_is_added_once(client):
    """Test mutator of a dataset is added only once to its creators list."""
    mutator = Person.from_git(client.repo)

    dataset = Dataset(
        name="my-dataset",
        creators=[mutator],
        date_published=datetime.datetime.now(datetime.timezone.utc),
        same_as="http://some-url",
    )
    old_dataset = copy.deepcopy(dataset)

    dataset.mutate()

    assert_dataset_is_mutated(old=old_dataset, new=dataset, mutator=mutator)
    assert 1 == len(dataset.creators)


@pytest.mark.xfail
def test_mutate_is_done_once():
    """Test dataset mutation can be done only once."""
    dataset = Dataset(name="my-dataset", creators=[])
    before_id = dataset._id

    dataset.mutate()
    after_id = dataset._id

    assert before_id != after_id

    dataset.mutate()
    assert after_id == dataset._id


@pytest.mark.parametrize(
    "name, slug",
    [
        ("UPPER-CASE", "upper-case"),
        (" a name  ", "a_name"),
        ("..non-alphanumeric start or end-_", "non-alphanumeric_start_or_end"),
        ("double ..__._- non-alphanumeric", "double_non-alphanumeric"),
        ("double..dots", "double.dots"),
        ("nön-àsçîï", "non-ascii"),
        ("ends in .lock", "ends_in_lock"),
    ],
)
def test_dataset_name_slug(name, slug):
    """Test slug generation from name."""
    assert slug == get_slug(name)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serialization tests for renku models."""

import datetime
from urllib.parse import urljoin

import pytest

from renku.core.management.migrations.models import v9 as old_datasets
from renku.core.utils.uuid import is_uuid


def test_dataset_deserialization(client_with_datasets, load_dataset_with_injection):
    """Test Dataset deserialization."""
    dataset = load_dataset_with_injection("dataset-1", client_with_datasets)

    dataset_types = {
        "date_created": [datetime.datetime],
        "creators": [list],
        "description": [str, type(None)],
        "files": [list],
        "identifier": [str],
        "keywords": [list],
    }

    for attribute, type_ in dataset_types.items():
        assert type(dataset.__getattribute__(attribute)) in type_

    creator_types = {"email": str, "id": str, "name": str, "affiliation": str}

    creator = load_dataset_with_injection("dataset-1", client_with_datasets).creators[0]

    for attribute, type_ in creator_types.items():
        assert type(getattr(creator, attribute)) is type_


@pytest.mark.xfail
def test_uuid_migration(dataset_metadata, client):
    """Test migration of id with UUID."""
    dataset = old_datasets.Dataset.from_jsonld(dataset_metadata)

    assert is_uuid(dataset.identifier)
    assert urljoin("/datasets/", dataset.identifier) == dataset.id


def test_dataset_creator_email(dataset_metadata):
    """Check that creators without an email are assigned a blank node."""
    # modify the dataset metadata to change the creator
    dataset = old_datasets.Dataset.from_jsonld(dataset_metadata)

    dataset.creators[0].id = "mailto:None"
    dataset_broken = old_datasets.Dataset.from_jsonld(dataset.as_jsonld())
    assert "mailto:None" not in dataset_broken.creators[0].id



# -*- coding: utf-8 -*-
#
# Copyright 2019-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dispatcher tests."""
from pathlib import Path

import pytest

from renku.core import errors
from renku.core.management import LocalClient
from renku.core.management.command_builder.client_dispatcher import ClientDispatcher
from renku.core.management.command_builder.database_dispatcher import DatabaseDispatcher


def test_client_dispatcher(tmpdir):
    """Test getting correct current client."""
    test_dir = tmpdir.mkdir("test")
    other_test_dir = tmpdir.mkdir("other_test")

    dispatcher = ClientDispatcher()

    with pytest.raises(errors.ConfigurationError):
        dispatcher.current_client

    dispatcher.push_client_to_stack(Path(tmpdir))

    assert dispatcher.current_client.path == Path(tmpdir).resolve()

    dispatcher.push_client_to_stack(str(Path(test_dir)))

    assert dispatcher.current_client.path == Path(test_dir).resolve()

    dispatcher.push_created_client_to_stack(LocalClient(Path(other_test_dir)))

    assert dispatcher.current_client.path == Path(other_test_dir).resolve()

    dispatcher.pop_client()

    assert dispatcher.current_client.path == Path(test_dir).resolve()

    dispatcher.pop_client()

    assert dispatcher.current_client.path == Path(tmpdir).resolve()

    dispatcher.pop_client()

    with pytest.raises(errors.ConfigurationError):
        dispatcher.current_client


def test_database_dispatcher(tmpdir):
    """Test getting correct database."""

    test_dir = tmpdir.mkdir("test")
    other_test_dir = tmpdir.mkdir("other_test")

    dispatcher = DatabaseDispatcher()

    with pytest.raises(errors.ConfigurationError):
        dispatcher.current_database

    dispatcher.push_database_to_stack(Path(tmpdir))

    assert dispatcher.current_database._storage.path == Path(tmpdir).resolve()

    dispatcher.push_database_to_stack(str(Path(test_dir)))

    assert dispatcher.current_database._storage.path == Path(test_dir).resolve()

    dispatcher.push_database_to_stack(Path(other_test_dir), commit=True)

    assert dispatcher.current_database._storage.path == Path(other_test_dir).resolve()

    dispatcher.pop_database()

    # NOTE: Make sure the database was committed on pop
    assert (Path(other_test_dir) / "root").exists()

    assert dispatcher.current_database._storage.path == Path(test_dir).resolve()

    dispatcher.pop_database()

    # NOTE: Make sure the database was not committed on pop
    assert not (Path(test_dir) / "root").exists()

    assert dispatcher.current_database._storage.path == Path(tmpdir).resolve()

    dispatcher.pop_database()

    # NOTE: Make sure the database was not committed on pop
    assert not (Path(tmpdir) / "root").exists()

    with pytest.raises(errors.ConfigurationError):
        dispatcher.current_database



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku CLI commands for handling of Knowledge Graph data.

Exporting Knowledge Graph data
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

You can export part or all of the Renku Knowledge Graph metadata for the
current project using the ``renku graph export`` command.

By default, this will export the metadata created in the last commit in the
project.
If that commit was not a ``renku`` command that creates metadata, it will
produce no output.

.. code-block:: console

   $ renku dataset create my-dataset
   OK
   $ renku graph export
    [
        {
            "@id": "https://localhost/datasets/850e74d6c0204e8c923457a1b9ce52d8",
            "@type": [
            "http://schema.org/Dataset",
            "http://www.w3.org/ns/prov#Entity"
            ],

            [... many more lines ...]

        }
    ]

Here we created a new dataset and then ``renku graph export`` exported the
created metadata as JSON-LD, the default format.

If you want the Knowledge Graph data for the whole project, you can use
``renku graph export --full``. Alternatively, you can get data for a single
commit by using ``renku graph export --revision <git commit sha>`` or by
specifying a range of commits like ``renku graph export --revision sha1..sha2``.

``renku graph export`` currently supports various formats for export, such as
``json-ld``, ``rdf``, ``nt`` (for triples) and ``dot`` (for GraphViz graphs),
which can be specified using the ``--format`` option. For instance,

.. code-block:: console

   $ renku graph export --full --format dot | dot -Tpng -o my_graph.png

would produce a PNG image of the whole Knowledge Graph for the project.

To run validation on the generated output, you can pass the ``--strict``
option, which will check that all the nodes and properties in the graph are
correct and that there isn't anything missing.

"""

import click

from renku.cli.utils.callback import ClickCallback
from renku.cli.utils.click import CaseInsensitiveChoice
from renku.core.commands.format.graph import GRAPH_FORMATS
from renku.core.commands.graph import export_graph_command


@click.group(hidden=True)
def graph():
    """Graph commands."""


@graph.command()
@click.option("--format", type=CaseInsensitiveChoice(GRAPH_FORMATS), default="json-ld", help="Choose an output format.")
@click.option(
    "--revision",
    type=str,
    default="HEAD",
    help="Limit graph to changes done in revision (or range of revisions like 'A..B').",
)
@click.option("--full", is_flag=True, help="Generate full graph for project. Overrides --revision.")
@click.option("--strict", is_flag=True, default=False, help="Validate triples before output.")
def export(format, revision, full, strict):
    r"""Export Renku graph metadata for project."""

    if full:
        revision = None

    communicator = ClickCallback()
    export_graph_command().with_communicator(communicator).build().execute(
        format=format, strict=strict, revision_or_range=revision
    )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test project database gateways."""

from datetime import datetime

from renku.core.metadata.gateway.project_gateway import ProjectGateway
from renku.core.models.project import Project
from renku.core.models.provenance.agent import Person


def test_project_gateway_update(dummy_database_injection_manager, monkeypatch):
    """Test updating project metadata."""

    project = Project(
        id=Project.generate_id("namespace", "my_project"),
        name="my_project",
        creator=Person(
            id=Person.generate_id("test@test.com", "Test tester"), email="test@test.com", name="Test tester"
        ),
        date_created=datetime.utcnow(),
    )

    with dummy_database_injection_manager(None):
        project_gateway = ProjectGateway()

        project_gateway.update_project(project)

        stored_project = project_gateway.get_project()

        assert stored_project.id == project.id
        assert stored_project.agent_version
        assert stored_project.name == "my_project"

        monkeypatch.setattr("renku.__version__", "999.999.999")

        stored_project.name = "new name"

        project_gateway.update_project(project)

        stored_project = project_gateway.get_project()

        assert stored_project.id == project.id
        assert stored_project.agent_version == "999.999.999"
        assert stored_project.name == "new name"
